#!/usr/bin/env python

"""
This script implements a homeostatic STDP rule for inhibitory
synapses onto excitatory neurons presented in Zenke et al. (2015) and
explained in detail in the accompanying supplementary information.

In essence, the magnitude and sign of the STDP kernel at inhibitory
synapses targeting excitatory neurons is modulated by a global factor
G, which corresponds to the difference between the average firing rate
of the excitatory postsynaptic neurons and a target firing rate
:math:`\gamma`. If the average firing rate of excitatory neurons exceeds the
target firing rate, pre/post pairings (and post/pre pairings) at
inhibitory synapses connecting to excitatory neurons result in
potentiation; depression of the inhibitory synapses occurs when the
excitatory neurons are unable to match the desired target rate.

The script emulates classic STDP protocols, simulating a connected
presynaptic and postsynaptic neuron pair with precisely controlled
spike times, and a large, independently firing population of
excitatory neurons.

Paul Brodersen, 2025

Reference:
Zenke, Friedemann, Everton J. Agnes, and Wulfram Gerstner.
"Diverse Synaptic Plasticity Mechanisms Orchestrated to Form and
Retrieve Memories in Spiking Neural Networks." Nature Communications 6,
no. 1 (April 21, 2015): 6922. https://doi.org/10.1038/ncomms7922.
"""

import numpy as np
import matplotlib.pyplot as plt
import brian2 as b2

net = b2.Network()

# excitatory neuron population
epoch_length = 1 * b2.second
spike_rate = np.arange(0, 12.5, 2.5)
total_epochs = len(spike_rate)
population_spike_rate = b2.TimedArray(spike_rate * b2.Hz, dt=epoch_length)
population_size = 1000
spike_generators = b2.PoissonGroup(population_size, rates="population_spike_rate(t)")
net.add(spike_generators)

# global factor G
gamma = 5 * b2.Hz
tau_H = 0.1 * b2.second # NB: value in Zenke et al (2015) is 10 seconds; reduced here to allow for shorter simulations
global_factor_model = """
G = H - gamma    : Hz
dH/dt = -H/tau_H : Hz
"""
global_factor = b2.NeuronGroup(1, model=global_factor_model)
net.add(global_factor)
spike_aggregator = b2.Synapses(spike_generators, global_factor, on_pre="H_post += 1/(tau_H * population_size)")
spike_aggregator.connect()
net.add(spike_aggregator)

# pre-synaptic neuron
spike_times = epoch_length * (np.arange(total_epochs) + 0.5)
spike_indices = np.zeros_like(spike_times / b2.second)
presynaptic_neuron = b2.SpikeGeneratorGroup(1, spike_indices, spike_times)
net.add(presynaptic_neuron)

# postsynaptic neuron
delay = 1 * b2.msecond
postsynaptic_neuron = b2.SpikeGeneratorGroup(1, spike_indices, spike_times + delay)
net.add(postsynaptic_neuron)

# synapse
tau_stdp = 20. * b2.msecond
eta = 1. * b2.nsiemens * b2.second
synapse_model = """
G : Hz (linked)
wij : siemens
dzi/dt = -zi / tau_stdp : 1 (event-driven)
dzj/dt = -zj / tau_stdp : 1 (event-driven)
"""
on_pre = """
zi += 1.
wij += eta * G * zj
"""
on_post = """
zj += 1.
wij += eta * G * (zi + 1)
"""
synapse = b2.Synapses(
    presynaptic_neuron, postsynaptic_neuron,
    model=synapse_model, on_pre=on_pre, on_post=on_post
)
synapse.connect(i=0, j=0)
synapse.G = b2.linked_var(global_factor, "G")
synapse.wij = 0 * b2.nsiemens
net.add(synapse)

# monitor
monitor = b2.StateMonitor(synapse, ["G", "wij"], record=True, dt=1*b2.msecond)
net.add(monitor)

# run
net.run(total_epochs * epoch_length)

# analysis
G = np.squeeze(monitor.G / b2.Hz)
wij = np.squeeze(monitor.wij / b2.nsiemens)
time = np.squeeze(monitor.t / b2.ms)
spike_times = spike_times / b2.ms

# visualize experimental protocol
fig, axes = plt.subplots(2, 1, sharex=True)
fig.suptitle("Experimental protocol")
axes[0].plot(time, G)
axes[0].set_ylabel("G [Hz]")
axes[1].plot(time, wij)
axes[1].set_ylabel("Weight [nS]")
for ax in axes:
    ax.vlines(spike_times, *ax.get_ylim(), color="gray", linestyle=":", label="Pre/post spike pairings")
axes[0].legend(loc="upper left")
axes[-1].set_xlabel("Time [ms]")

# plot relationship between G and \Delta wij
g = np.zeros_like(spike_times)
dw = np.zeros_like(spike_times)
for ii, spike_time in enumerate(spike_times):
    g[ii] = G[np.argmin(np.abs(time - spike_time))]
    w0 = wij[np.argmin(np.abs(time - spike_time - 10))]
    w1 = wij[np.argmin(np.abs(time - spike_time + 10))]
    dw[ii] = w1 - w0

fig, ax = plt.subplots()
ax.plot(g, dw, 'o')
ax.set_xlabel("G [Hz]")
ax.set_ylabel(r"$\Delta$weight [nS]")
ax.set_title("Weight change modulation")

plt.show()



import numbers
from collections.abc import Sequence

import numpy as np

from brian2.core.variables import Variables, get_dtype
from brian2.groups.group import CodeRunner, Group
from brian2.units.allunits import second
from brian2.units.fundamentalunits import Quantity
from brian2.utils.logger import get_logger

__all__ = ["StateMonitor"]

logger = get_logger(__name__)


class StateMonitorView:
    def __init__(self, monitor, item):
        self.monitor = monitor
        self.item = item
        self.indices = self._calc_indices(item)
        self._group_attribute_access_active = True

    def __getattr__(self, item):
        # We do this because __setattr__ and __getattr__ are not active until
        # _group_attribute_access_active attribute is set, and if it is set,
        # then __getattr__ will not be called. Therefore, if getattr is called
        # with this name, it is because it hasn't been set yet and so this
        # method should raise an AttributeError to agree that it hasn't been
        # called yet.
        if item == "_group_attribute_access_active":
            raise AttributeError
        if not hasattr(self, "_group_attribute_access_active"):
            raise AttributeError

        mon = self.monitor
        if item == "t":
            return Quantity(mon.variables["t"].get_value(), dim=second.dim)
        elif item == "t_":
            return mon.variables["t"].get_value()
        elif item in mon.record_variables:
            dims = mon.variables[item].dim
            return Quantity(
                mon.variables[item].get_value().T[self.indices], dim=dims, copy=True
            )
        elif item.endswith("_") and item[:-1] in mon.record_variables:
            return mon.variables[item[:-1]].get_value().T[self.indices].copy()
        else:
            raise AttributeError(f"Unknown attribute {item}")

    def _calc_indices(self, item):
        """
        Convert the neuron indices to indices into the stored values. For example, if neurons [0, 5, 10] have been
        recorded, [5, 10] is converted to [1, 2].
        """
        dtype = get_dtype(item)
        # scalar value
        if np.issubdtype(dtype, np.signedinteger) and not isinstance(item, np.ndarray):
            indices = np.nonzero(self.monitor.record == item)[0]
            if len(indices) == 0:
                raise IndexError(f"Index number {int(item)} has not been recorded")
            return indices[0]

        if self.monitor.record_all:
            return item
        indices = []
        for index in item:
            if index in self.monitor.record:
                indices.append(np.nonzero(self.monitor.record == index)[0][0])
            else:
                raise IndexError(f"Index number {int(index)} has not been recorded")
        return np.array(indices)

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"<{classname}, giving access to elements {self.item!r} recorded by "
            f"{self.monitor.name}>"
        )


class StateMonitor(Group, CodeRunner):
    """
    Record values of state variables during a run

    To extract recorded values after a run, use the ``t`` attribute for the
    array of times at which values were recorded, and variable name attribute
    for the values. The values will have shape ``(len(indices), len(t))``,
    where ``indices`` are the array indices which were recorded. When indexing
    the `StateMonitor` directly, the returned object can be used to get the
    recorded values for the specified indices, i.e. the indexing semantic
    refers to the indices in ``source``, not to the relative indices of the
    recorded values. For example, when recording only neurons with even numbers,
    `mon[[0, 2]].v` will return the values for neurons 0 and 2, whereas
    `mon.v[[0, 2]]` will return the values for the first and third *recorded*
    neurons, i.e. for neurons 0 and 4.

    Parameters
    ----------
    source : `Group`
        Which object to record values from.
    variables : str, sequence of str, True
        Which variables to record, or ``True`` to record all variables
        (note that this may use a great deal of memory).
    record : bool, sequence of ints
        Which indices to record, nothing is recorded for ``False``,
        everything is recorded for ``True`` (warning: may use a great deal of
        memory), or a specified subset of indices.
    dt : `Quantity`, optional
        The time step to be used for the monitor. Cannot be combined with
        the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the ``dt`` argument
        is specified, the clock of the `source` will be used.
    when : str, optional
        At which point during a time step the values should be recorded.
        Defaults to ``'start'``. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    name : str, optional
        A unique name for the object, otherwise will use
        ``source.name+'statemonitor_0'``, etc.
    codeobj_class : `CodeObject`, optional
        The `CodeObject` class to create.

    Examples
    --------

    Record all variables, first 5 indices::

        eqs = '''
        dV/dt = (2-V)/(10*ms) : 1
        '''
        threshold = 'V>1'
        reset = 'V = 0'
        G = NeuronGroup(100, eqs, threshold=threshold, reset=reset)
        G.V = rand(len(G))
        M = StateMonitor(G, True, record=range(5))
        run(100*ms)
        plot(M.t, M.V.T)
        show()

    Notes
    -----

    Since this monitor by default records in the ``'start'`` time slot,
    recordings of the membrane potential in integrate-and-fire models may look
    unexpected: the recorded membrane potential trace will never be above
    threshold in an integrate-and-fire model, because the reset statement will
    have been applied already. Set the ``when`` keyword to a different value if
    this is not what you want.

    Note that ``record=True`` only works in runtime mode for synaptic variables.
    This is because the actual array of indices has to be calculated and this is
    not possible in standalone mode, where the synapses have not been created
    yet at this stage. Consider using an explicit array of indices instead,
    i.e. something like ``record=np.arange(n_synapses)``.
    """

    invalidates_magic_network = False
    add_to_magic_network = True

    def __init__(
        self,
        source,
        variables,
        record,
        dt=None,
        clock=None,
        when="start",
        order=0,
        name="statemonitor*",
        codeobj_class=None,
    ):
        self.source = source
        # Make the monitor use the explicitly defined namespace of its source
        # group (if it exists)
        self.namespace = getattr(source, "namespace", None)
        self.codeobj_class = codeobj_class

        # run by default on source clock at the end
        if dt is None and clock is None:
            clock = source.clock

        # variables should always be a list of strings
        if variables is True:
            variables = source.equations.names
        elif isinstance(variables, str):
            variables = [variables]
        #: The variables to record
        self.record_variables = variables

        # record should always be an array of ints
        self.record_all = False
        if hasattr(record, "_indices"):
            # The ._indices method always returns absolute indices
            # If the source is already a subgroup of another group, we therefore
            # have to shift the indices to become relative to the subgroup
            record = record._indices() - getattr(source, "_offset", 0)
        if record is True:
            self.record_all = True
            try:
                record = np.arange(len(source), dtype=np.int32)
            except NotImplementedError:
                # In standalone mode, this is not possible for synaptic
                # variables because the number of synapses is not defined yet
                raise NotImplementedError(
                    "Cannot determine the actual "
                    "indices to record for record=True. "
                    "This can occur for example in "
                    "standalone mode when trying to "
                    "record a synaptic variable. "
                    "Consider providing an explicit "
                    "array of indices for the record "
                    "argument."
                )
        elif record is False:
            record = np.array([], dtype=np.int32)
        elif isinstance(record, numbers.Number):
            record = np.array([record], dtype=np.int32)
        else:
            record = np.asarray(record, dtype=np.int32)

        #: The array of recorded indices
        self.record = record
        self.n_indices = len(record)

        if not self.record_all:
            try:
                if len(record) and (
                    np.max(record) >= len(source) or np.min(record) < 0
                ):
                    # Check whether the values in record make sense
                    error_message = (
                        "The indices to record from contain values outside of the"
                        f" range [0, {len(source)-1}] allowed for the group"
                        f" '{source.name}'"
                    )
                    raise IndexError(error_message)
            except NotImplementedError:
                logger.warn(
                    "Cannot check whether the indices to record from are valid. This"
                    " can happen in standalone mode when recording from synapses that"
                    " have been created with a connection pattern. You can avoid this"
                    " situation by using synaptic indices in the connect call.",
                    name_suffix="cannot_check_statemonitor_indices",
                )

        # Some dummy code so that code generation takes care of the indexing
        # and subexpressions
        code = [f"_to_record_{v} = _source_{v}" for v in variables]
        code = "\n".join(code)

        CodeRunner.__init__(
            self,
            group=self,
            template="statemonitor",
            code=code,
            name=name,
            clock=clock,
            dt=dt,
            when=when,
            order=order,
            check_units=False,
        )

        self.add_dependency(source)

        # Setup variables
        self.variables = Variables(self)

        self.variables.add_dynamic_array(
            "t",
            size=0,
            dimensions=second.dim,
            constant=False,
            dtype=self._clock.variables["t"].dtype,
            read_only=True,
        )
        self.variables.add_array(
            "N", dtype=np.int32, size=1, scalar=True, read_only=True
        )
        self.variables.add_array(
            "_indices",
            size=len(self.record),
            dtype=self.record.dtype,
            constant=True,
            read_only=True,
            values=self.record,
        )
        self.variables.create_clock_variables(self._clock, prefix="_clock_")
        for varname in variables:
            var = source.variables[varname]
            if var.scalar and len(self.record) > 1:
                logger.warn(
                    "Variable %s is a shared variable but it will be "
                    "recorded once for every target." % varname,
                    once=True,
                )
            index = source.variables.indices[varname]
            self.variables.add_reference(
                f"_source_{varname}", source, varname, index=index
            )
            if index not in ("_idx", "0") and index not in variables:
                self.variables.add_reference(index, source)
            self.variables.add_dynamic_array(
                varname,
                size=(0, len(self.record)),
                resize_along_first=True,
                dimensions=var.dim,
                dtype=var.dtype,
                constant=False,
                read_only=True,
            )

        for varname in variables:
            var = self.source.variables[varname]
            self.variables.add_auxiliary_variable(
                f"_to_record_{varname}",
                dimensions=var.dim,
                dtype=var.dtype,
                scalar=var.scalar,
            )

        self.recorded_variables = {
            varname: self.variables[varname] for varname in variables
        }
        recorded_names = [varname for varname in variables]

        self.needed_variables = recorded_names
        self.template_kwds = {"_recorded_variables": self.recorded_variables}
        self.written_readonly_vars = {
            self.variables[varname] for varname in self.record_variables
        }
        self._enable_group_attributes()

    def resize(self, new_size):
        self.variables["N"].set_value(new_size)
        self.variables["t"].resize(new_size)

        for var in self.recorded_variables.values():
            var.resize((new_size, self.n_indices))

    def reinit(self):
        raise NotImplementedError()

    def __getitem__(self, item):
        dtype = get_dtype(item)
        if np.issubdtype(dtype, np.signedinteger):
            return StateMonitorView(self, item)
        elif isinstance(item, Sequence):
            index_array = np.array(item)
            if not np.issubdtype(index_array.dtype, np.signedinteger):
                raise TypeError("Index has to be an integer or a sequence of integers")
            return StateMonitorView(self, item)
        elif hasattr(item, "_indices"):
            # objects that support the indexing interface will return absolute
            # indices but here we need relative ones
            # TODO: How to we prevent the use of completely unrelated objects here?
            source_offset = getattr(self.source, "_offset", 0)
            return StateMonitorView(self, item._indices() - source_offset)
        else:
            raise TypeError(f"Cannot use object of type {type(item)} as an index")

    def __getattr__(self, item):
        # We do this because __setattr__ and __getattr__ are not active until
        # _group_attribute_access_active attribute is set, and if it is set,
        # then __getattr__ will not be called. Therefore, if getattr is called
        # with this name, it is because it hasn't been set yet and so this
        # method should raise an AttributeError to agree that it hasn't been
        # called yet.
        if item == "_group_attribute_access_active":
            raise AttributeError
        if not hasattr(self, "_group_attribute_access_active"):
            raise AttributeError
        if item in self.record_variables:
            var_dim = self.variables[item].dim
            return Quantity(self.variables[item].get_value().T, dim=var_dim, copy=True)
        elif item.endswith("_") and item[:-1] in self.record_variables:
            return self.variables[item[:-1]].get_value().T
        else:
            return Group.__getattr__(self, item)

    def __repr__(self):
        classname = self.__class__.__name__
        variables = repr(self.record_variables)
        return f"<{classname}, recording {variables} from '{self.source.name}'>"

    def record_single_timestep(self):
        """
        Records a single time step. Useful for recording the values at the end
        of the simulation -- otherwise a `StateMonitor` will not record the
        last simulated values since its ``when`` attribute defaults to
        ``'start'``, i.e. the last recording is at the *beginning* of the last
        time step.

        Notes
        -----
        This function will only work if the `StateMonitor` has been already run,
        but a run with a length of ``0*ms`` does suffice.

        Examples
        --------
        >>> from brian2 import *
        >>> G = NeuronGroup(1, 'dv/dt = -v/(5*ms) : 1')
        >>> G.v = 1
        >>> mon = StateMonitor(G, 'v', record=True)
        >>> run(0.5*ms)
        >>> print(np.array_str(mon.v[:], precision=3))
        [[ 1.     0.98   0.961  0.942  0.923]]
        >>> print(mon.t[:])
        [   0.  100.  200.  300.  400.] us
        >>> print(np.array_str(G.v[:], precision=3))  # last value had not been recorded
        [ 0.905]
        >>> mon.record_single_timestep()
        >>> print(mon.t[:])
        [   0.  100.  200.  300.  400.  500.] us
        >>> print(np.array_str(mon.v[:], precision=3))
        [[ 1.     0.98   0.961  0.942  0.923  0.905]]
        """
        if self.codeobj is None:
            raise TypeError(
                "Can only record a single time step after the "
                "network has been run once."
            )
        self.codeobj()



#!/usr/bin/env python
"""
Synfire chains
--------------
M. Diesmann et al. (1999). Stable propagation of synchronous spiking in cortical
neural networks. Nature 402, 529-533.
"""
from brian2 import *

duration = 100*ms

# Neuron model parameters
Vr = -70*mV
Vt = -55*mV
taum = 10*ms
taupsp = 0.325*ms
weight = 4.86*mV
# Neuron model
eqs = Equations('''
dV/dt = (-(V-Vr)+x)*(1./taum) : volt
dx/dt = (-x+y)*(1./taupsp) : volt
dy/dt = -y*(1./taupsp)+25.27*mV/ms+
        (39.24*mV/ms**0.5)*xi : volt
''')

# Neuron groups
n_groups = 10
group_size = 100
P = NeuronGroup(N=n_groups*group_size, model=eqs,
                threshold='V>Vt', reset='V=Vr', refractory=1*ms,
                method='euler')

Pinput = SpikeGeneratorGroup(85, np.arange(85),
                             np.random.randn(85)*1*ms + 50*ms)
# The network structure
S = Synapses(P, P, on_pre='y+=weight')
S.connect(j='k for k in range((int(i/group_size)+1)*group_size, (int(i/group_size)+2)*group_size) '
            'if i<N_pre-group_size')
Sinput = Synapses(Pinput, P[:group_size], on_pre='y+=weight')
Sinput.connect()

# Record the spikes
Mgp = SpikeMonitor(P)
Minput = SpikeMonitor(Pinput)
# Setup the network, and run it
P.V = 'Vr + rand() * (Vt - Vr)'
run(duration)

plot(Mgp.t/ms, 1.0*Mgp.i/group_size, '.')
plot([0, duration/ms], np.arange(n_groups).repeat(2).reshape(-1, 2).T, 'k-')
ylabel('group number')
yticks(np.arange(n_groups))
xlabel('time (ms)')
show()


"""
A cylinder plus two branches, with diameters according to Rall's formula
"""
from brian2 import *

defaultclock.dt = 0.01*ms

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV

# Morphology
diameter = 1*um
length = 300*um
Cm = 1*uF/cm**2
Ri = 150*ohm*cm
N = 500
rm = 1 / (gL * pi * diameter)  # membrane resistance per unit length
ra = (4 * Ri)/(pi * diameter**2)  # axial resistance per unit length
la = sqrt(rm / ra) # space length
morpho = Cylinder(diameter=diameter, length=length, n=N)
d1 = 0.5*um
L1 = 200*um
rm = 1 / (gL * pi * d1) # membrane resistance per unit length
ra = (4 * Ri) / (pi * d1**2) # axial resistance per unit length
l1 = sqrt(rm / ra) # space length
morpho.L = Cylinder(diameter=d1, length=L1, n=N)
d2 = (diameter**1.5 - d1**1.5)**(1. / 1.5)
rm = 1/(gL * pi * d2) # membrane resistance per unit length
ra = (4 * Ri) / (pi * d2**2) # axial resistance per unit length
l2 = sqrt(rm / ra) # space length
L2 = (L1 / l1) * l2
morpho.R = Cylinder(diameter=d2, length=L2, n=N)

eqs='''
Im = gL * (EL-v) : amp/meter**2
I : amp (point current)
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=Cm, Ri=Ri,
                       method='exponential_euler')
neuron.v = EL

neuron.I[0] = 0.02*nA # injecting at the left end
run(100*ms, report='text')

plot(neuron.main.distance/um, neuron.main.v/mV, 'k')
plot(neuron.L.distance/um, neuron.L.v/mV, 'k')
plot(neuron.R.distance/um, neuron.R.v/mV, 'k')
# Theory
x = neuron.main.distance
ra = la * 4 * Ri/(pi * diameter**2)
l = length/la + L1/l1
theory = EL + ra*neuron.I[0]*cosh(l - x/la)/sinh(l)
plot(x/um, theory/mV, 'r')
x = neuron.L.distance
theory = (EL+ra*neuron.I[0]*cosh(l - neuron.main.distance[-1]/la -
                                 (x - neuron.main.distance[-1])/l1)/sinh(l))
plot(x/um, theory/mV, 'r')
x = neuron.R.distance
theory = (EL+ra*neuron.I[0]*cosh(l - neuron.main.distance[-1]/la -
                                 (x - neuron.main.distance[-1])/l2)/sinh(l))
plot(x/um, theory/mV, 'r')
xlabel('x (um)')
ylabel('v (mV)')
show()



"""
Module handling equations and "code strings", expressions or statements, used
for example for the reset and threshold definition of a neuron.
"""

from .codestrings import Expression, Statements
from .equations import Equations

__all__ = ["Equations", "Expression", "Statements"]



"""
This model defines the `NeuronGroup`, the core of most simulations.
"""

import numbers
import string
from collections.abc import MutableMapping, Sequence

import numpy as np
import sympy
from pyparsing import Word

from brian2.codegen.translation import analyse_identifiers
from brian2.core.preferences import prefs
from brian2.core.spikesource import SpikeSource
from brian2.core.variables import Variables
from brian2.equations.equations import (
    DIFFERENTIAL_EQUATION,
    PARAMETER,
    SUBEXPRESSION,
    Equations,
    check_subexpressions,
    extract_constant_subexpressions,
)
from brian2.equations.refractory import add_refractoriness
from brian2.parsing.expressions import (
    is_boolean_expression,
    parse_expression_dimensions,
)
from brian2.stateupdaters.base import StateUpdateMethod
from brian2.units.allunits import second
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    Quantity,
    fail_for_dimension_mismatch,
)
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers

from .group import CodeRunner, Group, get_dtype
from .subgroup import Subgroup

__all__ = ["NeuronGroup"]

logger = get_logger(__name__)


IDENTIFIER = Word(
    f"{string.ascii_letters}_", f"{string.ascii_letters + string.digits}_"
).setResultsName("identifier")


def _valid_event_name(event_name):
    """
    Helper function to check whether a name is a valid name for an event.

    Parameters
    ----------
    event_name : str
        The name to check

    Returns
    -------
    is_valid : bool
        Whether the given name is valid
    """
    parse_result = list(IDENTIFIER.scanString(event_name))

    # parse_result[0][0][0] refers to the matched string -- this should be the
    # full identifier, if not it is an illegal identifier like "3foo" which only
    # matched on "foo"
    return len(parse_result) == 1 and parse_result[0][0][0] == event_name


def _guess_membrane_potential(equations):
    """
    Little helper function to guess which variable represents the membrane
    potential. This follows the same logic as in Brian1 but is only used to
    give a suggestion in the error message when a Brian1-style syntax is used
    for threshold or reset.
    """
    if len(equations) == 1:
        return list(equations.keys())[0]
    for name in equations:
        if name in ["V", "v", "Vm", "vm"]:
            return name

    # nothing found
    return None


# Note that we do not register this function with
# Equations.register_identifier_check, because we do not want this check to
# apply unconditionally to all equation objects ("x_post = ... : ... (summed)"
# needs to be allowed)
def check_identifier_pre_post(identifier):
    "Do not allow names ending in ``_pre`` or ``_post`` to avoid confusion."
    if identifier.endswith("_pre") or identifier.endswith("_post"):
        raise ValueError(
            f"'{identifier}' cannot be used as a variable name, the "
            "'_pre' and '_post' suffixes are used to refer to pre- and "
            "post-synaptic variables in synapses."
        )


def to_start_stop(item, N):
    """
    Helper function to transform a single number, a slice or an array of
    contiguous indices to a start and stop value. This is used to allow for
    some flexibility in the syntax of specifying subgroups in `.NeuronGroup`
    and `.SpatialNeuron`.

    Parameters
    ----------
    item : slice, int or sequence
        The slice, index, or sequence of indices to use. Note that a sequence
        of indices has to be a sorted ascending sequence of subsequent integers.
    N : int
        The total number of elements in the group.

    Returns
    -------
    start : int
        The start value of the slice.
    stop : int
        The stop value of the slice.

    Examples
    --------
    >>> from brian2.groups.neurongroup import to_start_stop
    >>> to_start_stop(slice(3, 6), 10)
    (3, 6)
    >>> to_start_stop(slice(3, None), 10)
    (3, 10)
    >>> to_start_stop(5, 10)
    (5, 6)
    >>> to_start_stop([3, 4, 5], 10)
    (3, 6)
    >>> to_start_stop([3, 5, 7], 10)
    Traceback (most recent call last):
        ...
    IndexError: Subgroups can only be constructed using contiguous indices.

    """
    if isinstance(item, slice):
        start, stop, step = item.indices(N)
    elif isinstance(item, numbers.Integral):
        start = item
        stop = item + 1
        step = 1
    elif isinstance(item, (Sequence, np.ndarray)) and not isinstance(item, str):
        if not (len(item) > 0 and np.all(np.diff(item) == 1)):
            raise IndexError(
                "Subgroups can only be constructed using contiguous indices."
            )
        if not np.issubdtype(np.asarray(item).dtype, np.integer):
            raise TypeError("Subgroups can only be constructed using integer values.")
        start = int(item[0])
        stop = int(item[-1]) + 1
        step = 1
    else:
        raise TypeError(
            "Subgroups can only be constructed using slicing "
            "syntax, a single index, or an array of contiguous "
            "indices."
        )
    if step != 1:
        raise IndexError("Subgroups have to be contiguous")
    if start >= stop:
        raise IndexError(
            f"Illegal start/end values for subgroup, {int(start)}>={int(stop)}"
        )
    if start >= N:
        raise IndexError(f"Illegal start value for subgroup, {int(start)}>={int(N)}")
    if stop > N:
        raise IndexError(f"Illegal stop value for subgroup, {int(stop)}>{int(N)}")
    if start < 0:
        raise IndexError("Indices have to be positive.")
    return start, stop


class StateUpdater(CodeRunner):
    """
    The `CodeRunner` that updates the state variables of a `NeuronGroup`
    at every timestep.
    """

    def __init__(self, group, method, method_options=None):
        self.method_choice = method
        self.method_options = method_options
        CodeRunner.__init__(
            self,
            group,
            "stateupdate",
            code="",  # will be set in update_abstract_code
            clock=group.clock,
            when="groups",
            order=group.order,
            name=f"{group.name}_stateupdater",
            check_units=False,
            generate_empty_code=False,
        )

    def _get_refractory_code(self, run_namespace):
        ref = self.group._refractory
        if ref is False:
            # No refractoriness
            abstract_code = ""
        elif isinstance(ref, Quantity):
            fail_for_dimension_mismatch(
                ref,
                second,
                "Refractory period has to "
                "be specified in units "
                "of seconds but got "
                "{value}",
                value=ref,
            )
            ref = float(ref)
            if prefs.legacy.refractory_timing:
                abstract_code = f"not_refractory = (t - lastspike) > {ref}\n"
            else:
                abstract_code = f"not_refractory = timestep(t - lastspike, dt) >= timestep({ref}, dt)\n"
        else:
            identifiers = get_identifiers(ref)
            variables = self.group.resolve_all(
                identifiers, run_namespace, user_identifiers=identifiers
            )
            dims = parse_expression_dimensions(str(ref), variables)
            if dims is second.dim:
                if prefs.legacy.refractory_timing:
                    abstract_code = f"(t - lastspike) > {ref}\n"
                else:
                    abstract_code = (
                        "not_refractory = timestep(t - lastspike, dt) >="
                        f" timestep({ref}, dt)\n"
                    )
            elif dims is DIMENSIONLESS:
                if not is_boolean_expression(str(ref), variables):
                    raise TypeError(
                        "Refractory expression is dimensionless "
                        "but not a boolean value. It needs to "
                        "either evaluate to a timespan or to a "
                        "boolean value."
                    )
                # boolean condition
                # we have to be a bit careful here, we can't just use the given
                # condition as it is, because we only want to *leave*
                # refractoriness, based on the condition
                abstract_code = f"not_refractory = not_refractory or not ({ref})\n"
            else:
                raise TypeError(
                    "Refractory expression has to evaluate to a "
                    "timespan or a boolean value, expression"
                    f"'{ref}' has units {dims} instead"
                )
        return abstract_code

    def update_abstract_code(self, run_namespace):
        # Update the not_refractory variable for the refractory period mechanism
        self.abstract_code = self._get_refractory_code(run_namespace=run_namespace)

        # Get the names used in the refractory code
        _, used_known, unknown = analyse_identifiers(
            self.abstract_code, self.group.variables, recursive=True
        )

        # Get all names used in the equations (and always get "dt")
        names = self.group.equations.names
        external_names = self.group.equations.identifiers | {"dt"}

        variables = self.group.resolve_all(
            used_known | unknown | names | external_names,
            run_namespace,
            # we don't need to raise any warnings
            # for the user here, warnings will
            # be raised in create_runner_codeobj
            user_identifiers=set(),
        )
        if len(self.group.equations.diff_eq_names) > 0:
            stateupdate_output = StateUpdateMethod.apply_stateupdater(
                self.group.equations,
                variables,
                self.method_choice,
                method_options=self.method_options,
                group_name=self.group.name,
            )
            if isinstance(stateupdate_output, str):
                self.abstract_code += stateupdate_output
            else:
                # Note that the reason to send self along with this method is so the StateUpdater
                # can be modified! i.e. in GSL StateUpdateMethod a custom CodeObject gets added
                # to the StateUpdater together with some auxiliary information
                self.abstract_code += stateupdate_output(self)

        user_code = "\n".join(
            [
                f"{var} = {expr}"
                for var, expr in self.group.equations.get_substituted_expressions(
                    variables
                )
            ]
        )
        self.user_code = user_code


class SubexpressionUpdater(CodeRunner):
    """
    The `CodeRunner` that updates the state variables storing the values of
    subexpressions that have been marked as "constant over dt".
    """

    def __init__(self, group, subexpressions, when="before_start"):
        code_lines = []
        for subexpr in subexpressions.ordered:
            code_lines.append(f"{subexpr.varname} = {subexpr.expr}")
        code = "\n".join(code_lines)
        CodeRunner.__init__(
            self,
            group,
            "stateupdate",
            code=code,  # will be set in update_abstract_code
            clock=group.clock,
            when=when,
            order=group.order,
            name=f"{group.name}_subexpression_update*",
        )


class Thresholder(CodeRunner):
    """
    The `CodeRunner` that applies the threshold condition to the state
    variables of a `NeuronGroup` at every timestep and sets its ``spikes``
    and ``refractory_until`` attributes.
    """

    def __init__(self, group, when="thresholds", event="spike"):
        self.event = event
        if group._refractory is False or event != "spike":
            template_kwds = {"_uses_refractory": False}
            needed_variables = []
        else:
            template_kwds = {"_uses_refractory": True}
            needed_variables = ["t", "not_refractory", "lastspike"]
        # Since this now works for general events not only spikes, we have to
        # pass the information about which variable to use to the template,
        # it can not longer simply refer to "_spikespace"
        eventspace_name = f"_{event}space"
        template_kwds["eventspace_variable"] = group.variables[eventspace_name]
        needed_variables.append(eventspace_name)
        self.variables = Variables(self)
        self.variables.add_auxiliary_variable("_cond", dtype=bool)
        CodeRunner.__init__(
            self,
            group,
            "threshold",
            code="",  # will be set in update_abstract_code
            clock=group.clock,
            when=when,
            order=group.order,
            name=f"{group.name}_{event}_thresholder",
            needed_variables=needed_variables,
            template_kwds=template_kwds,
        )

    def update_abstract_code(self, run_namespace):
        code = self.group.events[self.event]
        # Raise a useful error message when the user used a Brian1 syntax
        if not isinstance(code, str):
            if isinstance(code, Quantity):
                t = "a quantity"
            else:
                t = f"{type(code)}"
            error_msg = f"Threshold condition has to be a string, not {t}."
            if self.event == "spike":
                try:
                    vm_var = _guess_membrane_potential(self.group.equations)
                except AttributeError:  # not a group with equations...
                    vm_var = None
                if vm_var is not None:
                    error_msg += f" Probably you intended to use '{vm_var} > ...'?"
            raise TypeError(error_msg)

        self.user_code = f"_cond = {code}"

        identifiers = get_identifiers(code)
        variables = self.group.resolve_all(
            identifiers, run_namespace, user_identifiers=identifiers
        )
        if not is_boolean_expression(code, variables):
            raise TypeError(f"Threshold condition '{code}' is not a boolean expression")
        if self.group._refractory is False or self.event != "spike":
            self.abstract_code = f"_cond = {code}"
        else:
            self.abstract_code = f"_cond = ({code}) and not_refractory"


class Resetter(CodeRunner):
    """
    The `CodeRunner` that applies the reset statement(s) to the state
    variables of neurons that have spiked in this timestep.
    """

    def __init__(self, group, when="resets", order=None, event="spike"):
        self.event = event
        # Since this now works for general events not only spikes, we have to
        # pass the information about which variable to use to the template,
        # it can not longer simply refer to "_spikespace"
        eventspace_name = f"_{event}space"
        template_kwds = {"eventspace_variable": group.variables[eventspace_name]}
        needed_variables = [eventspace_name]
        order = order if order is not None else group.order
        CodeRunner.__init__(
            self,
            group,
            "reset",
            code="",  # will be set in update_abstract_code
            clock=group.clock,
            when=when,
            order=order,
            name=f"{group.name}_{event}_resetter",
            override_conditional_write=["not_refractory"],
            needed_variables=needed_variables,
            template_kwds=template_kwds,
        )

    def update_abstract_code(self, run_namespace):
        code = self.group.event_codes[self.event]
        # Raise a useful error message when the user used a Brian1 syntax
        if not isinstance(code, str):
            if isinstance(code, Quantity):
                t = "a quantity"
            else:
                t = f"{type(code)}"
            error_msg = f"Reset statement has to be a string, not {t}."
            if self.event == "spike":
                vm_var = _guess_membrane_potential(self.group.equations)
                if vm_var is not None:
                    error_msg += f" Probably you intended to use '{vm_var} = ...'?"
            raise TypeError(error_msg)

        self.abstract_code = code


class NeuronGroup(Group, SpikeSource):
    """
    A group of neurons.


    Parameters
    ----------
    N : int
        Number of neurons in the group.
    model : str, `Equations`
        The differential equations defining the group
    method : (str, function), optional
        The numerical integration method. Either a string with the name of a
        registered method (e.g. "euler") or a function that receives an
        `Equations` object and returns the corresponding abstract code. If no
        method is specified, a suitable method will be chosen automatically.
    threshold : str, optional
        The condition which produces spikes. Should be a single line boolean
        expression.
    reset : str, optional
        The (possibly multi-line) string with the code to execute on reset.
    refractory : {str, `Quantity`}, optional
        Either the length of the refractory period (e.g. ``2*ms``), a string
        expression that evaluates to the length of the refractory period
        after each spike (e.g. ``'(1 + rand())*ms'``), or a string expression
        evaluating to a boolean value, given the condition under which the
        neuron stays refractory after a spike (e.g. ``'v > -20*mV'``)
    events : dict, optional
        User-defined events in addition to the "spike" event defined by the
        ``threshold``. Has to be a mapping of strings (the event name) to
        strings (the condition) that will be checked.
    namespace: dict, optional
        A dictionary mapping identifier names to objects. If not given, the
        namespace will be filled in at the time of the call of `Network.run`,
        with either the values from the ``namespace`` argument of the
        `Network.run` method or from the local context, if no such argument is
        given.
    dtype : (`dtype`, `dict`), optional
        The `numpy.dtype` that will be used to store the values, or a
        dictionary specifying the type for variable names. If a value is not
        provided for a variable (or no value is provided at all), the preference
        setting `core.default_float_dtype` is used.
    codeobj_class : class, optional
        The `CodeObject` class to run code with.
    dt : `Quantity`, optional
        The time step to be used for the simulation. Cannot be combined with
        the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    name : str, optional
        A unique name for the group, otherwise use ``neurongroup_0``, etc.

    Notes
    -----
    `NeuronGroup` contains a `StateUpdater`, `Thresholder` and `Resetter`, and
    these are run at the 'groups', 'thresholds' and 'resets' slots (i.e. the
    values of their `when` attribute take these values). The `order`
    attribute will be passed down to the contained objects but can be set
    individually by setting the `order` attribute of the `state_updater`,
    `thresholder` and `resetter` attributes, respectively.
    """

    add_to_magic_network = True

    def __init__(
        self,
        N,
        model,
        method=("exact", "euler", "heun"),
        method_options=None,
        threshold=None,
        reset=None,
        refractory=False,
        events=None,
        namespace=None,
        dtype=None,
        dt=None,
        clock=None,
        order=0,
        name="neurongroup*",
        codeobj_class=None,
    ):
        Group.__init__(
            self,
            dt=dt,
            clock=clock,
            when="start",
            order=order,
            namespace=namespace,
            name=name,
        )
        if dtype is None:
            dtype = {}
        if isinstance(dtype, MutableMapping):
            dtype["lastspike"] = self._clock.variables["t"].dtype

        self.codeobj_class = codeobj_class

        try:
            self._N = N = int(N)
        except ValueError:
            if isinstance(N, str):
                raise TypeError(
                    "First NeuronGroup argument should be size, not equations."
                )
            raise
        if N < 1:
            raise ValueError(f"NeuronGroup size should be at least 1, was {str(N)}")

        self.start = 0
        self.stop = self._N

        ##### Prepare and validate equations
        if isinstance(model, str):
            model = Equations(model)
        if not isinstance(model, Equations):
            raise TypeError(
                "model has to be a string or an Equations "
                f"object, is '{type(model)}' instead."
            )

        # Check flags
        model.check_flags(
            {
                DIFFERENTIAL_EQUATION: ("unless refractory",),
                PARAMETER: ("constant", "shared", "linked"),
                SUBEXPRESSION: ("shared", "constant over dt"),
            }
        )

        # add refractoriness
        #: The original equations as specified by the user (i.e. without
        #: the multiplied `int(not_refractory)` term for equations marked as
        #: `(unless refractory)`)
        self.user_equations = model
        if refractory is not False:
            model = add_refractoriness(model)
        uses_refractoriness = len(model) and any(
            [
                "unless refractory" in eq.flags
                for eq in model.values()
                if eq.type == DIFFERENTIAL_EQUATION
            ]
        )

        # Separate subexpressions depending whether they are considered to be
        # constant over a time step or not
        model, constant_over_dt = extract_constant_subexpressions(model)
        self.equations = model

        self._linked_variables = set()
        logger.diagnostic(
            f"Creating NeuronGroup of size {self._N}, equations {self.equations}."
        )

        # All of the following will be created in before_run

        #: The refractory condition or timespan
        self._refractory = refractory
        if uses_refractoriness and refractory is False:
            logger.warn(
                'Model equations use the "unless refractory" flag but '
                "no refractory keyword was given.",
                "no_refractory",
            )

        #: The state update method selected by the user
        self.method_choice = method

        if events is None:
            events = {}

        if threshold is not None and (reset is None and refractory is False):
            if not ("rand(" in threshold or "randn(" in threshold):
                logger.warn(
                    f"The NeuronGroup '{self.name}' sets a threshold but "
                    "neither a reset condition nor a refractory "
                    "condition has been set. Did you forget either of "
                    "those? If this was intended, set the reset "
                    "argument to an empty string in order to avoid "
                    "this warning.",
                    name_suffix="only_threshold",
                )

        if threshold is not None:
            if "spike" in events:
                raise ValueError(
                    "The NeuronGroup defines both a threshold and a 'spike' event"
                )
            events["spike"] = threshold

        # Setup variables
        # Since we have to create _spikespace and possibly other "eventspace"
        # variables, we pass the supported events
        self._create_variables(dtype, events=list(events.keys()))

        #: Events supported by this group
        self.events = events

        #: Code that is triggered on events (e.g. reset)
        self.event_codes = {}

        #: Checks the spike threshold (or abitrary user-defined events)
        self.thresholder = {}

        #: Reset neurons which have spiked (or perform arbitrary actions for
        #: user-defined events)
        self.resetter = {}

        for event_name in events.keys():
            if not isinstance(event_name, str):
                raise TypeError(
                    "Keys in the 'events' dictionary have to be "
                    f"strings, not type {event_name}."
                )
            if not _valid_event_name(event_name):
                raise TypeError(
                    f"The name '{event_name}' cannot be used as an event name."
                )
            # By default, user-defined events are checked after the threshold
            when = "thresholds" if event_name == "spike" else "after_thresholds"
            # creating a Thresholder will take care of checking the validity
            # of the condition
            thresholder = Thresholder(self, event=event_name, when=when)
            self.thresholder[event_name] = thresholder
            self.contained_objects.append(thresholder)

        if reset is not None:
            self.run_on_event("spike", reset, when="resets")

        #: Performs numerical integration step
        self.state_updater = StateUpdater(self, method, method_options)
        self.contained_objects.append(self.state_updater)

        #: Update the "constant over a time step" subexpressions
        self.subexpression_updater = None
        if len(constant_over_dt):
            self.subexpression_updater = SubexpressionUpdater(self, constant_over_dt)
            self.contained_objects.append(self.subexpression_updater)

        if refractory is not False:
            # Set the refractoriness information
            self.variables["lastspike"].set_value(-1e4 * second)
            self.variables["not_refractory"].set_value(True)

        # Activate name attribute access
        self._enable_group_attributes()

    @property
    def spikes(self):
        """
        The spikes returned by the most recent thresholding operation.
        """
        # Note that we have to directly access the ArrayVariable object here
        # instead of using the Group mechanism by accessing self._spikespace
        # Using the latter would cut _spikespace to the length of the group
        spikespace = self.variables["_spikespace"].get_value()
        return spikespace[: spikespace[-1]]

    def state(self, name, use_units=True, level=0):
        try:
            return Group.state(self, name, use_units=use_units, level=level + 1)
        except KeyError as ex:
            if name in self._linked_variables:
                raise TypeError(f"Link target for variable {name} has not been set.")
            else:
                raise ex

    def run_on_event(self, event, code, when="after_resets", order=None):
        """
        Run code triggered by a custom-defined event (see `NeuronGroup`
        documentation for the specification of events).The created `Resetter`
        object will be automatically added to the group, it therefore does not
        need to be added to the network manually. However, a reference to the
        object will be returned, which can be used to later remove it from the
        group or to set it to inactive.

        Parameters
        ----------
        event : str
            The name of the event that should trigger the code
        code : str
            The code that should be executed
        when : str, optional
            The scheduling slot that should be used to execute the code.
            Defaults to `'after_resets'`. See :ref:`scheduling` for possible values.
        order : int, optional
            The order for operations in the same scheduling slot. Defaults to
            the order of the `NeuronGroup`.

        Returns
        -------
        obj : `Resetter`
            A reference to the object that will be run.
        """
        if event not in self.events:
            error_message = f"Unknown event '{event}'."
            if event == "spike":
                error_message += " Did you forget to define a threshold?"
            raise ValueError(error_message)
        if event in self.resetter:
            raise ValueError(
                "Cannot add code for event '%s', code for this "
                "event has already been added." % event
            )
        self.event_codes[event] = code
        resetter = Resetter(self, when=when, order=order, event=event)
        self.resetter[event] = resetter
        self.contained_objects.append(resetter)

        return resetter

    def set_event_schedule(self, event, when="after_thresholds", order=None):
        """
        Change the scheduling slot for checking the condition of an event.

        Parameters
        ----------
        event : str
            The name of the event for which the scheduling should be changed
        when : str, optional
            The scheduling slot that should be used to check the condition.
            Defaults to `'after_thresholds'`. See :ref:`scheduling` for possible values.
        order : int, optional
            The order for operations in the same scheduling slot. Defaults to
            the order of the `NeuronGroup`.
        """
        if event not in self.thresholder:
            raise ValueError(f"Unknown event '{event}'.")
        order = order if order is not None else self.order
        self.thresholder[event].when = when
        self.thresholder[event].order = order

    def __getitem__(self, item):
        start, stop = to_start_stop(item, self._N)

        return Subgroup(self, start, stop)

    def _create_variables(self, user_dtype, events):
        """
        Create the variables dictionary for this `NeuronGroup`, containing
        entries for the equation variables and some standard entries.
        """
        self.variables = Variables(self)
        self.variables.add_constant("N", self._N)

        # Standard variables always present
        for event in events:
            self.variables.add_array(
                f"_{event}space", size=self._N + 1, dtype=np.int32, constant=False
            )
        # Add the special variable "i" which can be used to refer to the neuron index
        self.variables.add_arange("i", size=self._N, constant=True, read_only=True)
        # Add the clock variables
        self.variables.create_clock_variables(self._clock)

        for eq in self.equations.values():
            dtype = get_dtype(eq, user_dtype)
            check_identifier_pre_post(eq.varname)
            if eq.type in (DIFFERENTIAL_EQUATION, PARAMETER):
                if "linked" in eq.flags:
                    # 'linked' cannot be combined with other flags
                    if not len(eq.flags) == 1:
                        raise SyntaxError(
                            "The 'linked' flag cannot be combined with other flags"
                        )
                    self._linked_variables.add(eq.varname)
                else:
                    constant = "constant" in eq.flags
                    shared = "shared" in eq.flags
                    size = 1 if shared else self._N
                    self.variables.add_array(
                        eq.varname,
                        size=size,
                        dimensions=eq.dim,
                        dtype=dtype,
                        constant=constant,
                        scalar=shared,
                    )
            elif eq.type == SUBEXPRESSION:
                self.variables.add_subexpression(
                    eq.varname,
                    dimensions=eq.dim,
                    expr=str(eq.expr),
                    dtype=dtype,
                    scalar="shared" in eq.flags,
                )
            else:
                raise AssertionError(f"Unknown type of equation: {eq.eq_type}")

        # Add the conditional-write attribute for variables with the
        # "unless refractory" flag
        if self._refractory is not False:
            for eq in self.equations.values():
                if eq.type == DIFFERENTIAL_EQUATION and "unless refractory" in eq.flags:
                    not_refractory_var = self.variables["not_refractory"]
                    var = self.variables[eq.varname]
                    var.set_conditional_write(not_refractory_var)

        # Stochastic variables
        for xi in self.equations.stochastic_variables:
            self.variables.add_auxiliary_variable(xi, dimensions=(second**-0.5).dim)

        # Check scalar subexpressions
        for eq in self.equations.values():
            if eq.type == SUBEXPRESSION and "shared" in eq.flags:
                var = self.variables[eq.varname]
                for identifier in var.identifiers:
                    if identifier in self.variables:
                        if not self.variables[identifier].scalar:
                            raise SyntaxError(
                                f"Shared subexpression '{eq.varname}' "
                                "refers to non-shared variable "
                                f"'{identifier}'."
                            )

    def before_run(self, run_namespace=None):
        # Check units
        self.equations.check_units(self, run_namespace=run_namespace)
        # Check that subexpressions that refer to stateful functions are labeled
        # as "constant over dt"
        check_subexpressions(self, self.equations, run_namespace)
        super().before_run(run_namespace=run_namespace)

    def _repr_html_(self):
        text = [rf"NeuronGroup '{self.name}' with {self._N} neurons.<br>"]
        text.append(r"<b>Model:</b><nr>")
        text.append(sympy.latex(self.equations))

        def add_event_to_text(event):
            if event == "spike":
                event_header = "Spiking behaviour"
                event_condition = "Threshold condition"
                event_code = "Reset statement(s)"
            else:
                event_header = f'Event "{event}"'
                event_condition = "Event condition"
                event_code = "Executed statement(s)"
            condition = self.events[event]
            text.append(
                rf'<b>{event_header}:</b><ul style="list-style-type: none; margin-top:'
                r' 0px;">'
            )
            text.append(rf"<li><i>{event_condition}: </i>")
            text.append(f"<code>{str(condition)}</code></li>")
            statements = self.event_codes.get(event, None)
            if statements is not None:
                text.append(rf"<li><i>{event_code}:</i>")
                if "\n" in str(statements):
                    text.append("</br>")
                text.append(rf"<code>{str(statements)}</code></li>")
            text.append("</ul>")

        if "spike" in self.events:
            add_event_to_text("spike")
        for event in self.events:
            if event != "spike":
                add_event_to_text(event)

        return "\n".join(text)



#!/usr/bin/env python
"""
Jeffress model, adapted with spiking neuron models.
A sound source (white noise) is moving around the head.
Delay differences between the two ears are used to determine the azimuth of the
source. Delays are mapped to a neural place code using delay lines (each neuron
receives input from both ears, with different delays).
"""
from brian2 import *

defaultclock.dt = .02*ms

# Sound
sound = TimedArray(10 * randn(50000), dt=defaultclock.dt) # white noise

# Ears and sound motion around the head (constant angular speed)
sound_speed = 300*metre/second
interaural_distance = 20*cm # big head!
max_delay = interaural_distance / sound_speed
print("Maximum interaural delay: %s" % max_delay)
angular_speed = 2 * pi / second # 1 turn/second
tau_ear = 1*ms
sigma_ear = .1
eqs_ears = '''
dx/dt = (sound(t-delay)-x)/tau_ear+sigma_ear*(2./tau_ear)**.5*xi : 1 (unless refractory)
delay = distance*sin(theta) : second
distance : second # distance to the centre of the head in time units
dtheta/dt = angular_speed : radian
'''
ears = NeuronGroup(2, eqs_ears, threshold='x>1', reset='x = 0',
                   refractory=2.5*ms, name='ears', method='euler')
ears.distance = [-.5 * max_delay, .5 * max_delay]
traces = StateMonitor(ears, 'delay', record=True)
# Coincidence detectors
num_neurons = 30
tau = 1*ms
sigma = .1
eqs_neurons = '''
dv/dt = -v / tau + sigma * (2 / tau)**.5 * xi : 1
'''
neurons = NeuronGroup(num_neurons, eqs_neurons, threshold='v>1',
                      reset='v = 0', name='neurons', method='euler')

synapses = Synapses(ears, neurons, on_pre='v += .5')
synapses.connect()

synapses.delay['i==0'] = '(1.0*j)/(num_neurons-1)*1.1*max_delay'
synapses.delay['i==1'] = '(1.0*(num_neurons-j-1))/(num_neurons-1)*1.1*max_delay'

spikes = SpikeMonitor(neurons)

run(1000*ms)

# Plot the results
i, t = spikes.it
subplot(2, 1, 1)
plot(t/ms, i, '.')
xlabel('Time (ms)')
ylabel('Neuron index')
xlim(0, 1000)
subplot(2, 1, 2)
plot(traces.t/ms, traces.delay.T/ms)
xlabel('Time (ms)')
ylabel('Input delay (ms)')
xlim(0, 1000)
tight_layout()
show()



"""
An example that uses a function from external C library (OpenCV in this case).
Works for all C-based code generation targets (i.e. for cython and cpp_standalone
device) and for numpy (using the Python bindings).

This example needs a working installation of OpenCV 3.x and its Python bindings.
It has been tested on 64 bit Linux in a conda environment with packages from the
``conda-forge`` channels (opencv 3.4.4, x264 1!152.20180717, ffmpeg 4.1).
"""
import os
import urllib.request, urllib.error, urllib.parse
import cv2  # Import OpenCV2

from brian2 import *

defaultclock.dt = 1*ms
prefs.codegen.target = 'cython'
prefs.logging.std_redirection = False
set_device('cpp_standalone', clean=True)
filename = os.path.abspath('Megamind.avi')

if not os.path.exists(filename):
    print('Downloading the example video file')
    response = urllib.request.urlopen('http://docs.opencv.org/2.4/_downloads/Megamind.avi')
    data = response.read()
    with open(filename, 'wb') as f:
        f.write(data)

video = cv2.VideoCapture(filename)
width, height, frame_count = (int(video.get(cv2.CAP_PROP_FRAME_WIDTH)),
                              int(video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
                              int(video.get(cv2.CAP_PROP_FRAME_COUNT)))
fps = 24
time_between_frames = 1*second/fps

@implementation('cpp', '''
double* get_frame(bool new_frame)
{
    // The following initializations will only be executed once
    static cv::VideoCapture source("VIDEO_FILENAME");
    static cv::Mat frame;
    static double* grayscale_frame = (double*)malloc(VIDEO_WIDTH*VIDEO_HEIGHT*sizeof(double));
    if (new_frame)
    {
        source >> frame;
        double mean_value = 0;
        for (int row=0; row<VIDEO_HEIGHT; row++)
            for (int col=0; col<VIDEO_WIDTH; col++)
            {
                const double grayscale_value = (frame.at<cv::Vec3b>(row, col)[0] +
                                                frame.at<cv::Vec3b>(row, col)[1] +
                                                frame.at<cv::Vec3b>(row, col)[2])/(3.0*128);
                mean_value += grayscale_value / (VIDEO_WIDTH * VIDEO_HEIGHT);
                grayscale_frame[row*VIDEO_WIDTH + col] = grayscale_value;
            }
        // subtract the mean
        for (int i=0; i<VIDEO_HEIGHT*VIDEO_WIDTH; i++)
            grayscale_frame[i] -= mean_value;
    }
    return grayscale_frame;
}

double video_input(const int x, const int y)
{
    // Get the current frame (or a new frame in case we are asked for the first
    // element
    double *frame = get_frame(x==0 && y==0);
    return frame[y*VIDEO_WIDTH + x];
}
'''.replace('VIDEO_FILENAME', filename),
                libraries=['opencv_core',
                           'opencv_highgui',
                           'opencv_videoio'],
                headers=['<opencv2/core/core.hpp>',
                         '<opencv2/highgui/highgui.hpp>'],
                define_macros=[('VIDEO_WIDTH', width),
                               ('VIDEO_HEIGHT', height)])
@check_units(x=1, y=1, result=1)
def video_input(x, y):
    # we assume this will only be called in the custom operation (and not for
    # example in a reset or synaptic statement), so we don't need to do indexing
    # but we can directly return the full result
    _, frame = video.read()
    grayscale = frame.mean(axis=2)
    grayscale /= 128.  # scale everything between 0 and 2
    return grayscale.ravel() - grayscale.ravel().mean()


N = width * height
tau, tau_th = 10*ms, time_between_frames
G = NeuronGroup(N, '''dv/dt = (-v + I)/tau : 1
                      dv_th/dt = -v_th/tau_th : 1
                      row : integer (constant)
                      column : integer (constant)
                      I : 1 # input current''',
                threshold='v>v_th', reset='v=0; v_th = 3*v_th + 1.0',
                method='exact')
G.v_th = 1
G.row = 'i//width'
G.column = 'i%width'

G.run_regularly('I = video_input(column, row)',
                dt=time_between_frames)
mon = SpikeMonitor(G)
runtime = frame_count*time_between_frames
run(runtime, report='text')

# Avoid going through the whole Brian2 indexing machinery too much
i, t, row, column = mon.i[:], mon.t[:], G.row[:], G.column[:]

import matplotlib.animation as animation

# TODO: Use overlapping windows
stepsize = 100*ms
def next_spikes():
    step = next_spikes.step
    if step*stepsize > runtime:
        next_spikes.step=0
        raise StopIteration()
    spikes = i[(t>=step*stepsize) & (t<(step+1)*stepsize)]
    next_spikes.step += 1
    yield column[spikes], row[spikes]
next_spikes.step = 0

fig, ax = plt.subplots()
dots, = ax.plot([], [], 'k.', markersize=2, alpha=.25)
ax.set_xlim(0, width)
ax.set_ylim(0, height)
ax.invert_yaxis()
def run(data):
    x, y = data
    dots.set_data(x, y)

ani = animation.FuncAnimation(fig, run, next_spikes, blit=False, repeat=True,
                              repeat_delay=1000)
plt.show()



#!/usr/bin/env python
"""
Spike-timing dependent plasticity.
Adapted from Song, Miller and Abbott (2000) and Song and Abbott (2001).

This example is modified from ``synapses_STDP.py`` and writes a standalone
C++ project in the directory ``STDP_standalone``.
"""
from brian2 import *

set_device('cpp_standalone', directory='STDP_standalone')

N = 1000
taum = 10*ms
taupre = 20*ms
taupost = taupre
Ee = 0*mV
vt = -54*mV
vr = -60*mV
El = -74*mV
taue = 5*ms
F = 15*Hz
gmax = .01
dApre = .01
dApost = -dApre * taupre / taupost * 1.05
dApost *= gmax
dApre *= gmax

eqs_neurons = '''
dv/dt = (ge * (Ee-v) + El - v) / taum : volt
dge/dt = -ge / taue : 1
'''

input = PoissonGroup(N, rates=F)
neurons = NeuronGroup(1, eqs_neurons, threshold='v>vt', reset='v = vr',
                      method='euler')
S = Synapses(input, neurons,
             '''w : 1
                dApre/dt = -Apre / taupre : 1 (event-driven)
                dApost/dt = -Apost / taupost : 1 (event-driven)''',
             on_pre='''ge += w
                    Apre += dApre
                    w = clip(w + Apost, 0, gmax)''',
             on_post='''Apost += dApost
                     w = clip(w + Apre, 0, gmax)''',
             )
S.connect()
S.w = 'rand() * gmax'
mon = StateMonitor(S, 'w', record=[0, 1])
s_mon = SpikeMonitor(input)

run(100*second, report='text')

subplot(311)
plot(S.w / gmax, '.k')
ylabel('Weight / gmax')
xlabel('Synapse index')
subplot(312)
hist(S.w / gmax, 20)
xlabel('Weight / gmax')
subplot(313)
plot(mon.t/second, mon.w.T/gmax)
xlabel('Time (s)')
ylabel('Weight / gmax')
tight_layout()
show()



import os
import platform
import tempfile

import pytest
from numpy.testing import assert_allclose, assert_equal

from brian2 import *
from brian2.devices import device_module
from brian2.devices.device import reinit_and_delete, reset_device, set_device
from brian2.tests.utils import assert_allclose
from brian2.utils.logger import catch_logs


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_cpp_standalone():
    set_device("cpp_standalone", build_on_run=False)
    ##### Define the model
    tau = 1 * ms
    eqs = """
    dV/dt = (-40*mV-V)/tau : volt (unless refractory)
    """
    threshold = "V>-50*mV"
    reset = "V=-60*mV"
    refractory = 5 * ms
    N = 1000

    G = NeuronGroup(
        N, eqs, reset=reset, threshold=threshold, refractory=refractory, name="gp"
    )
    G.V = "-i*mV"
    M = SpikeMonitor(G)
    S = Synapses(G, G, "w : volt", on_pre="V += w")
    S.connect("abs(i-j)<5 and i!=j")
    S.w = 0.5 * mV
    S.delay = "0*ms"

    net = Network(G, M, S)
    net.run(100 * ms)
    device.build(directory=None, with_output=False)
    # we do an approximate equality here because depending on minor details of how it was compiled, the results
    # may be slightly different (if -ffast-math is on)
    assert len(M.i) >= 17000 and len(M.i) <= 18000
    assert len(M.t) == len(M.i)
    assert M.t[0] == 0.0
    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_multiple_connects():
    set_device("cpp_standalone", build_on_run=False)
    G = NeuronGroup(10, "v:1")
    S = Synapses(G, G, "w:1")
    S.connect(i=[0], j=[0])
    S.connect(i=[1], j=[1])
    run(0 * ms)
    device.build(directory=None, with_output=False)
    assert len(S) == 2 and len(S.w[:]) == 2
    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_device_cache_synapses():
    # Check that we can ask for known synaptic information at runtime
    set_device("cpp_standalone", build_on_run=False)
    G = NeuronGroup(10, "v:1")
    S = Synapses(G, G, "w:1")
    S.connect(i=[0], j=[0])
    assert len(S) == 1
    assert_equal(S.i[:], [0])
    assert_equal(S.j[:], [0])

    S.connect(i=[1], j=[1])
    assert len(S) == 2
    assert_equal(S.i[:], [0, 1])
    assert_equal(S.j[:], [0, 1])

    S.connect(p=0.1)  # We can't know anything about synapses anymore

    with pytest.raises(NotImplementedError):
        len(S)

    with pytest.raises(NotImplementedError):
        S.i[:]

    with pytest.raises(NotImplementedError):
        S.j[:]

    S.connect(i=[1], j=[1])
    # Synapses are still "unknown" due to the previous p=0.1 call
    with pytest.raises(NotImplementedError):
        len(S)

    with pytest.raises(NotImplementedError):
        S.i[:]

    with pytest.raises(NotImplementedError):
        S.j[:]


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_storing_loading():
    set_device("cpp_standalone", build_on_run=False)
    G = NeuronGroup(
        10,
        """
        v : volt
        x : 1
        n : integer
        b : boolean
        """,
    )
    v = np.arange(10) * volt
    x = np.arange(10, 20)
    n = np.arange(20, 30)
    b = np.array([True, False]).repeat(5)
    G.v = v
    G.x = x
    G.n = n
    G.b = b
    S = Synapses(
        G,
        G,
        """
        v_syn : volt
        x_syn : 1
        n_syn : integer
        b_syn : boolean
        """,
    )
    S.connect(j="i")
    S.v_syn = v
    S.x_syn = x
    S.n_syn = n
    S.b_syn = b
    run(0 * ms)
    device.build(directory=None, with_output=False)
    assert_allclose(G.v[:], v)
    assert_allclose(S.v_syn[:], v)
    assert_allclose(G.x[:], x)
    assert_allclose(S.x_syn[:], x)
    assert_allclose(G.n[:], n)
    assert_allclose(S.n_syn[:], n)
    assert_allclose(G.b[:], b)
    assert_allclose(S.b_syn[:], b)
    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
@pytest.mark.openmp
def test_openmp_consistency():
    previous_device = get_device()
    n_cells = 100
    n_recorded = 10
    numpy.random.seed(42)
    taum = 20 * ms
    taus = 5 * ms
    Vt = -50 * mV
    Vr = -60 * mV
    El = -49 * mV
    fac = 60 * 0.27 / 10
    gmax = 20 * fac
    dApre = 0.01
    taupre = 20 * ms
    taupost = taupre
    dApost = -dApre * taupre / taupost * 1.05
    dApost *= 0.1 * gmax
    dApre *= 0.1 * gmax

    connectivity = numpy.random.randn(n_cells, n_cells)
    sources = numpy.random.randint(0, n_cells - 1, 10 * n_cells)
    # Only use one spike per time step (to rule out that a single source neuron
    # has more than one spike in a time step)
    times = (
        numpy.random.choice(numpy.arange(10 * n_cells), 10 * n_cells, replace=False)
        * ms
    )
    v_init = Vr + numpy.random.rand(n_cells) * (Vt - Vr)

    eqs = Equations(
        """
        dv/dt = (g-(v-El))/taum : volt
        dg/dt = -g/taus         : volt
        """
    )

    results = {}

    for n_threads, devicename in [
        (0, "runtime"),
        (0, "cpp_standalone"),
        (1, "cpp_standalone"),
        (2, "cpp_standalone"),
        (3, "cpp_standalone"),
        (4, "cpp_standalone"),
    ]:
        set_device(devicename, build_on_run=False, with_output=False)
        Synapses.__instances__().clear()
        if devicename == "cpp_standalone":
            reinit_and_delete()
        prefs.devices.cpp_standalone.openmp_threads = n_threads
        P = NeuronGroup(
            n_cells, model=eqs, threshold="v>Vt", reset="v=Vr", refractory=5 * ms
        )
        Q = SpikeGeneratorGroup(n_cells, sources, times)
        P.v = v_init
        P.g = 0 * mV
        S = Synapses(
            P,
            P,
            model="""
            dApre/dt=-Apre/taupre    : 1 (event-driven)    
            dApost/dt=-Apost/taupost : 1 (event-driven)
            w                        : 1
            """,
            pre="""
            g     += w*mV
            Apre  += dApre
            w      = w + Apost
            """,
            post="""
            Apost += dApost
            w      = w + Apre
            """,
        )
        S.connect()

        S.w = fac * connectivity.flatten()

        T = Synapses(Q, P, model="w : 1", on_pre="g += w*mV")
        T.connect(j="i")
        T.w = 10 * fac

        spike_mon = SpikeMonitor(P)
        rate_mon = PopulationRateMonitor(P)
        state_mon = StateMonitor(S, "w", record=np.arange(n_recorded), dt=0.1 * second)
        v_mon = StateMonitor(P, "v", record=np.arange(n_recorded))

        run(0.2 * second, report="text")

        if devicename == "cpp_standalone":
            device.build(directory=None, with_output=False)

        results[n_threads, devicename] = {}
        results[n_threads, devicename]["w"] = state_mon.w
        results[n_threads, devicename]["v"] = v_mon.v
        results[n_threads, devicename]["s"] = spike_mon.num_spikes
        results[n_threads, devicename]["r"] = rate_mon.rate[:]

    for key1, key2 in [
        ((0, "runtime"), (0, "cpp_standalone")),
        ((1, "cpp_standalone"), (0, "cpp_standalone")),
        ((2, "cpp_standalone"), (0, "cpp_standalone")),
        ((3, "cpp_standalone"), (0, "cpp_standalone")),
        ((4, "cpp_standalone"), (0, "cpp_standalone")),
    ]:
        assert_allclose(results[key1]["w"], results[key2]["w"])
        assert_allclose(results[key1]["v"], results[key2]["v"])
        assert_allclose(results[key1]["r"], results[key2]["r"])
        assert_allclose(results[key1]["s"], results[key2]["s"])
    reset_device(previous_device)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_duplicate_names_across_nets():
    set_device("cpp_standalone", build_on_run=False)
    # In standalone mode, names have to be globally unique, not just unique
    # per network
    obj1 = BrianObject(name="name1")
    obj2 = BrianObject(name="name2")
    obj3 = BrianObject(name="name3")
    obj4 = BrianObject(name="name1")
    net1 = Network(obj1, obj2)
    net2 = Network(obj3, obj4)
    net1.run(0 * ms)
    net2.run(0 * ms)
    with pytest.raises(ValueError):
        device.build()

    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
@pytest.mark.openmp
def test_openmp_scalar_writes():
    # Test that writing to a scalar variable only is done once in an OpenMP
    # setting (see github issue #551)
    set_device("cpp_standalone", build_on_run=False)
    prefs.devices.cpp_standalone.openmp_threads = 4
    G = NeuronGroup(10, "s : 1 (shared)")
    G.run_regularly("s += 1")
    run(defaultclock.dt)
    device.build(directory=None, with_output=False)
    assert_equal(G.s[:], 1.0)

    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_time_after_run():
    set_device("cpp_standalone", build_on_run=False)
    # Check that the clock and network time after a run is correct, even if we
    # have not actually run the code yet (via build)
    G = NeuronGroup(10, "dv/dt = -v/(10*ms) : 1")
    net = Network(G)
    assert_allclose(defaultclock.dt, 0.1 * ms)
    assert_allclose(defaultclock.t, 0.0 * ms)
    assert_allclose(G.t, 0.0 * ms)
    assert_allclose(net.t, 0.0 * ms)
    net.run(10 * ms)
    assert_allclose(defaultclock.t, 10.0 * ms)
    assert_allclose(G.t, 10.0 * ms)
    assert_allclose(net.t, 10.0 * ms)
    net.run(10 * ms)
    assert_allclose(defaultclock.t, 20.0 * ms)
    assert_allclose(G.t, 20.0 * ms)
    assert_allclose(net.t, 20.0 * ms)
    device.build(directory=None, with_output=False)
    # Everything should of course still be accessible
    assert_allclose(defaultclock.t, 20.0 * ms)
    assert_allclose(G.t, 20.0 * ms)
    assert_allclose(net.t, 20.0 * ms)

    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_array_cache():
    # Check that variables are only accessible from Python when they should be
    set_device("cpp_standalone", build_on_run=False)
    G = NeuronGroup(
        10,
        """
        dv/dt = -v / (10*ms) : 1
        w : 1
        x : 1
        y : 1
        z : 1 (shared)
        """,
        threshold="v>1",
    )
    S = Synapses(G, G, "weight: 1", on_pre="w += weight")
    S.connect(p=0.2)
    S.weight = 7
    # All neurongroup values should be known
    assert_allclose(G.v, 0)
    assert_allclose(G.w, 0)
    assert_allclose(G.x, 0)
    assert_allclose(G.y, 0)
    assert_allclose(G.z, 0)
    assert_allclose(G.i, np.arange(10))

    # But the synaptic variable is not -- we don't know the number of synapses
    with pytest.raises(NotImplementedError):
        S.weight[:]

    # Setting variables with explicit values should not change anything
    G.v = np.arange(10) + 1
    G.w = 2
    G.y = 5
    G.z = 7
    assert_allclose(G.v, np.arange(10) + 1)
    assert_allclose(G.w, 2)
    assert_allclose(G.y, 5)
    assert_allclose(G.z, 7)

    # But setting with code should invalidate them
    G.x = "i*2"
    with pytest.raises(NotImplementedError):
        G.x[:]

    # Make sure that the array cache does not allow to use incorrectly sized
    # values to pass
    with pytest.raises(ValueError):
        setattr(G, "w", [0, 2])
    with pytest.raises(ValueError):
        G.w.__setitem__(slice(0, 4), [0, 2])

    run(10 * ms)
    # v is now no longer known without running the network
    with pytest.raises(NotImplementedError):
        G.v[:]
    # Neither is w, it is updated in the synapse
    with pytest.raises(NotImplementedError):
        G.w[:]
    # However, no code touches y or z
    assert_allclose(G.y, 5)
    assert_allclose(G.z, 7)
    # i is read-only anyway
    assert_allclose(G.i, np.arange(10))

    # After actually running the network, everything should be accessible
    device.build(directory=None, with_output=False)
    assert all(G.v > 0)
    assert all(G.w > 0)
    assert_allclose(G.x, np.arange(10) * 2)
    assert_allclose(G.y, 5)
    assert_allclose(G.z, 7)
    assert_allclose(G.i, np.arange(10))
    assert_allclose(S.weight, 7)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_run_with_debug():
    # We just want to make sure that it works for now (i.e. not fails with a
    # compilation or runtime error), capturing the output is actually
    # a bit involved to get right.
    set_device("cpp_standalone", build_on_run=True, debug=True, directory=None)
    group = NeuronGroup(1, "v: 1", threshold="False")
    syn = Synapses(group, group, on_pre="v += 1")
    syn.connect()
    mon = SpikeMonitor(group)
    run(defaultclock.dt)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_run_with_synapses_and_profile():
    set_device("cpp_standalone", build_on_run=True, directory=None)
    group = NeuronGroup(1, "v: 1", threshold="False", reset="")
    syn = Synapses(group, group, on_pre="v += 1")
    syn.connect()
    mon = SpikeMonitor(group)
    run(defaultclock.dt, profile=True)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_changing_profile_arg():
    set_device("cpp_standalone", build_on_run=False)
    G = NeuronGroup(10000, "v : 1")
    op1 = G.run_regularly("v = exp(-v)", name="op1")
    op2 = G.run_regularly("v = exp(-v)", name="op2")
    op3 = G.run_regularly("v = exp(-v)", name="op3")
    op4 = G.run_regularly("v = exp(-v)", name="op4")
    # Op 1 is active only during the first profiled run
    # Op 2 is active during both profiled runs
    # Op 3 is active only during the second profiled run
    # Op 4 is never active (only during the unprofiled run)
    op1.active = True
    op2.active = True
    op3.active = False
    op4.active = False
    run(1000 * defaultclock.dt, profile=True)
    op1.active = True
    op2.active = True
    op3.active = True
    op4.active = True
    run(1000 * defaultclock.dt, profile=False)
    op1.active = False
    op2.active = True
    op3.active = True
    op4.active = False
    run(1000 * defaultclock.dt, profile=True)
    device.build(directory=None, with_output=False)
    profiling_dict = dict(magic_network.profiling_info)
    # Note that for now, C++ standalone creates a new CodeObject for every run,
    # which is most of the time unnecessary (this is partly due to the way we
    # handle constants: they are included as literals in the code but they can
    # change between runs). Therefore, the profiling info is potentially
    # difficult to interpret
    assert len(profiling_dict) == 4  # 2 during first run, 2 during last run
    # The two code objects that were executed during the first run
    assert (
        "op1_codeobject" in profiling_dict
        and profiling_dict["op1_codeobject"] > 0 * second
    )
    assert (
        "op2_codeobject" in profiling_dict
        and profiling_dict["op2_codeobject"] > 0 * second
    )
    # Four code objects were executed during the second run, but no profiling
    # information was saved
    for name in [
        "op1_codeobject_1",
        "op2_codeobject_1",
        "op3_codeobject",
        "op4_codeobject",
    ]:
        assert name not in profiling_dict
    # Two code objects were exectued during the third run
    assert (
        "op2_codeobject_2" in profiling_dict
        and profiling_dict["op2_codeobject_2"] > 0 * second
    )
    assert (
        "op3_codeobject_1" in profiling_dict
        and profiling_dict["op3_codeobject_1"] > 0 * second
    )


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_profile_via_set_device_arg():
    set_device("cpp_standalone", build_on_run=False, profile=True)
    G = NeuronGroup(10000, "v : 1")
    op1 = G.run_regularly("v = exp(-v)", name="op1")
    op2 = G.run_regularly("v = exp(-v)", name="op2")
    op3 = G.run_regularly("v = exp(-v)", name="op3")
    op4 = G.run_regularly("v = exp(-v)", name="op4")
    # Op 1 is active only during the first profiled run
    # Op 2 is active during both profiled runs
    # Op 3 is active only during the second profiled run
    # Op 4 is never active (only during the unprofiled run)
    op1.active = True
    op2.active = True
    op3.active = False
    op4.active = False
    run(1000 * defaultclock.dt)  # Should use profile=True via set_device
    op1.active = True
    op2.active = True
    op3.active = True
    op4.active = True
    run(1000 * defaultclock.dt, profile=False)
    op1.active = False
    op2.active = True
    op3.active = True
    op4.active = False
    run(1000 * defaultclock.dt, profile=True)
    device.build(directory=None, with_output=False)
    profiling_dict = dict(magic_network.profiling_info)
    # Note that for now, C++ standalone creates a new CodeObject for every run,
    # which is most of the time unnecessary (this is partly due to the way we
    # handle constants: they are included as literals in the code but they can
    # change between runs). Therefore, the profiling info is potentially
    # difficult to interpret
    assert len(profiling_dict) == 4  # 2 during first run, 2 during last run
    # The two code objects that were executed during the first run
    assert (
        "op1_codeobject" in profiling_dict
        and profiling_dict["op1_codeobject"] > 0 * second
    )
    assert (
        "op2_codeobject" in profiling_dict
        and profiling_dict["op2_codeobject"] > 0 * second
    )
    # Four code objects were executed during the second run, but no profiling
    # information was saved
    for name in [
        "op1_codeobject_1",
        "op2_codeobject_1",
        "op3_codeobject",
        "op4_codeobject",
    ]:
        assert name not in profiling_dict
    # Two code objects were exectued during the third run
    assert (
        "op2_codeobject_2" in profiling_dict
        and profiling_dict["op2_codeobject_2"] > 0 * second
    )
    assert (
        "op3_codeobject_1" in profiling_dict
        and profiling_dict["op3_codeobject_1"] > 0 * second
    )


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_delete_code_data():
    set_device("cpp_standalone", build_on_run=False)
    group = NeuronGroup(10, "dv/dt = -v / (10*ms) : volt", method="exact")
    group.v = np.arange(10) * mV  # uses the static array mechanism
    run(defaultclock.dt)
    # Overwrite the initial values via run_args mechanims
    device.build(run_args={group.v: np.arange(10)[::-1] * mV}, directory=None)
    results_dir = os.path.join(device.project_dir, "results")
    assert os.path.exists(results_dir) and os.path.isdir(results_dir)
    # There should be 3 files for the clock, 2 for the neurongroup (index + v),
    # and the "last_run_info.txt" file
    assert len(os.listdir(results_dir)) == 6
    device.delete(data=True, run_args=False, code=False, directory=False)
    assert os.path.exists(results_dir) and os.path.isdir(results_dir)
    assert len(os.listdir(results_dir)) == 0
    static_arrays_before = len(
        os.listdir(os.path.join(device.project_dir, "static_arrays"))
    )
    assert static_arrays_before > 0
    assert len(os.listdir(os.path.join(device.project_dir, "code_objects"))) > 0
    device.delete(data=False, code=True, run_args=False, directory=False)
    assert (
        0
        < len(os.listdir(os.path.join(device.project_dir, "static_arrays")))
        < static_arrays_before
    )
    assert len(os.listdir(os.path.join(device.project_dir, "code_objects"))) == 0
    device.delete(data=False, code=False, run_args=True, directory=False)
    len(os.listdir(os.path.join(device.project_dir, "static_arrays"))) == 0


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_delete_directory():
    set_device("cpp_standalone", build_on_run=True, directory=None)
    group = NeuronGroup(10, "dv/dt = -v / (10*ms) : volt", method="exact")
    group.v = np.arange(10) * mV  # uses the static array mechanism
    run(defaultclock.dt)
    # Add a new file
    dummy_file = os.path.join(device.project_dir, "results", "dummy.txt")
    open(dummy_file, "w").flush()
    assert os.path.isfile(dummy_file)
    with catch_logs() as logs:
        device.delete(directory=True)
    assert (
        len(
            [
                l
                for l in logs
                if l[1] == "brian2.devices.cpp_standalone.device.delete_skips_directory"
            ]
        )
        == 1
    )
    assert os.path.isfile(dummy_file)
    with catch_logs() as logs:
        device.delete(directory=True, force=True)
    assert (
        len(
            [
                l
                for l in logs
                if l[1] == "brian2.devices.cpp_standalone.device.delete_skips_directory"
            ]
        )
        == 0
    )
    # everything should be deleted
    assert not os.path.exists(device.project_dir)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_multiple_standalone_runs():
    # see github issue #1189
    set_device("cpp_standalone", directory=None)

    network = Network()
    Pe = NeuronGroup(1, "v : 1", threshold="False")
    C_ee = Synapses(Pe, Pe, on_pre="v += 1")
    C_ee.connect()
    network.add(Pe, C_ee)
    network.run(defaultclock.dt)

    device.reinit()
    device.activate(directory=None)

    network2 = Network()
    Pe = NeuronGroup(1, "v : 1", threshold="False")
    C_ee = Synapses(Pe, Pe, on_pre="v += 1")
    C_ee.connect()
    network2.add(Pe, C_ee)
    network2.run(defaultclock.dt)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_continued_standalone_runs():
    # see github issue #1237
    set_device("cpp_standalone", build_on_run=False)

    source = SpikeGeneratorGroup(1, [0], [0] * ms)
    target = NeuronGroup(1, "v : 1")
    C_ee = Synapses(source, target, on_pre="v += 1", delay=2 * ms)
    C_ee.connect()
    run(1 * ms)
    # Spike has not been delivered yet
    run(2 * ms)

    device.build(directory=None)
    assert target.v[0] == 1  # Make sure the spike got delivered


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_constant_replacement():
    # see github issue #1276
    set_device("cpp_standalone")
    x = 42
    G = NeuronGroup(1, "y : 1")
    G.y = "x"
    run(0 * ms)
    assert G.y[0] == 42.0


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_change_parameter_without_recompile():
    prefs.core.default_float_dtype = np.float32
    set_device("cpp_standalone", directory=None, with_output=True, debug=True)
    on_off = TimedArray([True, False, True], dt=defaultclock.dt, name="on_off")
    stim = TimedArray(
        np.arange(30).reshape(3, 10) * nA, dt=defaultclock.dt, name="stim"
    )
    G = NeuronGroup(
        10,
        """
        x : 1 (constant)
        v : volt (constant)
        n : integer (constant)
        b : boolean (constant)
        s = int(on_off(t))*stim(t, i) : amp
        """,
        name="neurons",
    )
    G.x = np.arange(10)
    G.n = np.arange(10)
    G.b = np.arange(10) % 2 == 0
    G.v = np.arange(10) * volt
    mon = StateMonitor(G, "s", record=True)
    run(3 * defaultclock.dt)
    assert array_equal(G.x, np.arange(10))
    assert array_equal(G.n, np.arange(10))
    assert array_equal(G.b, np.arange(10) % 2 == 0)
    assert array_equal(G.v, np.arange(10) * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # on_off(t) == False
                [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
            ]
        ),
    )

    device.run(
        run_args=[
            "neurons.x=5",
            "neurons.v=3",
            "neurons.n=17",
            "neurons.b=True",
            "on_off.values=True",
        ]
    )
    assert array_equal(G.x, np.ones(10) * 5)
    assert array_equal(G.n, np.ones(10) * 17)
    assert array_equal(G.b, np.ones(10, dtype=bool))
    assert array_equal(G.v, np.ones(10) * 3 * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
            ]
        ),
    )
    ar = np.arange(10) * 2.0
    ar.astype(G.x.dtype).tofile(os.path.join(device.project_dir, "init_values_x1.dat"))
    ar.astype(G.n.dtype).tofile(os.path.join(device.project_dir, "init_values_n1.dat"))
    (np.arange(10) % 2 != 0).tofile(
        os.path.join(device.project_dir, "init_values_b1.dat")
    )
    ar.astype(G.v.dtype).tofile(os.path.join(device.project_dir, "init_values_v1.dat"))
    ar2 = 2 * np.arange(30).reshape(3, 10) * nA
    ar2.astype(stim.values.dtype).tofile(
        os.path.join(device.project_dir, "init_stim_values.dat")
    )
    device.run(
        run_args=[
            "neurons.v=init_values_v1.dat",
            "neurons.x=init_values_x1.dat",
            "neurons.b=init_values_b1.dat",
            "neurons.n=init_values_n1.dat",
            "stim.values=init_stim_values.dat",
        ]
    )
    assert array_equal(G.x, ar)
    assert array_equal(G.n, ar)
    assert array_equal(G.b, np.arange(10) % 2 != 0)
    assert array_equal(G.v, ar * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 2, 4, 6, 8, 10, 12, 14, 16, 18],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # on_off(t) == False
                [40, 42, 44, 46, 48, 50, 52, 54, 56, 58],
            ]
        ),
    )
    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_change_parameter_without_recompile_errors():
    set_device("cpp_standalone", directory=None, with_output=False)
    G = NeuronGroup(10, "v:volt", name="neurons")
    G.v = np.arange(10) * volt

    run(0 * ms)

    with pytest.raises(DimensionMismatchError):
        device.run(run_args={G.v: 5})
    with pytest.raises(DimensionMismatchError):
        device.run(run_args={G.v: 5 * siemens})
    with pytest.raises(TypeError):
        device.run(run_args={G.v: np.arange(9) * volt})

    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_change_parameter_without_recompile_dict_syntax():
    set_device("cpp_standalone", directory=None, with_output=False)
    on_off = TimedArray([True, False, True], dt=defaultclock.dt, name="on_off")
    stim = TimedArray(
        np.arange(30).reshape(3, 10) * nA, dt=defaultclock.dt, name="stim"
    )
    G = NeuronGroup(
        10,
        """
        x : 1 (constant)
        n : integer (constant)
        b : boolean (constant)
        v : volt (constant)
        s = int(on_off(t))*stim(t, i) : amp
        """,
        name="neurons",
    )
    G.x = np.arange(10)
    G.n = np.arange(10)
    G.b = np.arange(10) % 2 == 0
    G.v = np.arange(10) * volt
    mon = StateMonitor(G, "s", record=True)
    run(3 * defaultclock.dt)
    assert array_equal(G.x, np.arange(10))
    assert array_equal(G.n, np.arange(10))
    assert array_equal(G.b, np.arange(10) % 2 == 0)
    assert array_equal(G.v, np.arange(10) * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # on_off(t) == False
                [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
            ]
        ),
    )
    device.run(run_args={G.x: 5, G.v: 3 * volt, G.n: 17, G.b: True, on_off: True})
    assert array_equal(G.x, np.ones(10) * 5)
    assert array_equal(G.n, np.ones(10) * 17)
    assert array_equal(G.b, np.ones(10, dtype=bool))
    assert array_equal(G.v, np.ones(10) * 3 * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
            ]
        ),
    )
    ar = np.arange(10) * 2.0
    ar2 = 2 * np.arange(30).reshape(3, 10) * nA
    device.run(
        run_args={
            G.x: ar,
            G.v: ar * volt,
            G.n: ar,
            G.b: np.arange(10) % 2 != 0,
            stim: ar2,
        }
    )
    assert array_equal(G.x, ar)
    assert array_equal(G.n, ar)
    assert array_equal(G.b, np.arange(10) % 2 != 0)
    assert array_equal(G.v, ar * volt)
    assert_allclose(
        mon.s.T / nA,
        np.array(
            [
                [0, 2, 4, 6, 8, 10, 12, 14, 16, 18],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # on_off(t) == False
                [40, 42, 44, 46, 48, 50, 52, 54, 56, 58],
            ]
        ),
    )
    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_change_synapse_parameter_without_recompile_dict_syntax():
    set_device("cpp_standalone", directory=None, with_output=False)
    G = NeuronGroup(10, "", name="neurons")
    S = Synapses(G, G, "w:1", name="Synapses")
    S.connect(j="i")
    S.w = np.arange(10)
    run(0 * ms)
    assert array_equal(S.w, np.arange(10))

    device.run(run_args={S.w: 17})
    assert array_equal(S.w, np.ones(10) * 17)

    ar = np.arange(10) * 2.0
    device.run(run_args={S.w: ar})
    assert array_equal(S.w, ar)

    reset_device()


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_change_parameter_without_recompile_dependencies():
    set_device("cpp_standalone", directory=None, with_output=False)
    G = NeuronGroup(
        10,
        """
        v:volt
        w:1
        """,
        name="neurons",
    )
    G.v = np.arange(10) * volt
    device.apply_run_args()
    G.w = "v/volt*2"

    run(0 * ms)
    assert array_equal(G.v, np.arange(10))
    assert array_equal(G.w, np.arange(10) * 2)

    device.run(run_args=["neurons.v=5"])
    assert array_equal(G.v, np.ones(10) * 5 * volt)
    assert array_equal(G.w, np.ones(10) * 5 * 2)

    ar = np.arange(10) * 2.0
    ar.astype(G.v.dtype).tofile(os.path.join(device.project_dir, "init_values_v2.dat"))
    device.run(run_args=[f"neurons.v=init_values_v2.dat"])
    assert array_equal(G.v, ar * volt)
    assert array_equal(G.w, ar * 2)

    reset_device()


class RunSim:
    def __init__(self):
        self.device = get_device()
        self.G = NeuronGroup(
            10,
            """
            v:volt
            w:1
            x:1
            """,
            name="neurons",
        )
        run(0 * ms)

    def run_sim(self, idx):
        # Ugly hack needed for windows
        device_module.active_device = self.device
        device.run(
            results_directory=f"results_{idx}",
            run_args={
                self.G.v: idx * volt,
                self.G.w: np.arange(10),  # Same values for all processes
                self.G.x: np.arange(10) * idx,  # Different values
            },
        )
        return self.G.v[:], self.G.w[:], self.G.x[:]


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
@pytest.mark.skipif(
    platform.system() == "Darwin" and platform.processor() == "arm",
    reason="multiprocessing hangs on macOS with Apple Silicon",
)
def test_change_parameters_multiprocessing():
    set_device("cpp_standalone", directory=None)
    sim = RunSim()

    import multiprocessing

    p = multiprocessing.Pool()
    try:
        results = p.map(sim.run_sim, range(5))
    finally:
        p.close()
        p.join()

    for idx, result in zip(range(5), results):
        v, w, x = result
        assert array_equal(v, np.ones(10) * idx * volt)
        assert array_equal(w, np.arange(10))
        assert array_equal(x, np.arange(10) * idx)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_header_file_inclusion():
    set_device("cpp_standalone", directory=None, debug=True)
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, "foo.h"), "w") as f:
            f.write(
                """
            namespace brian_test_namespace {
                extern double test_variable;
            }
            """
            )
        with open(os.path.join(tmpdir, "foo.cpp"), "w") as f:
            f.write(
                """
            namespace brian_test_namespace {
                double test_variable = 42;
            }
            """
            )

        @implementation(
            "cpp",
            """
        double brian_function(int index) {
            using namespace brian_test_namespace;
            return test_variable * index;
        }
        """,
            headers=['"foo.h"'],
            sources=[os.path.join(tmpdir, "foo.cpp")],
            include_dirs=[tmpdir],
        )
        @check_units(index=1, result=1)
        def brian_function(index):
            raise NotImplementedError()

        # Use the function in a somewhat convoluted way that exposes errors in the
        # code generation process
        G = PoissonGroup(5, rates="brian_function(i)*Hz")
        S = Synapses(G, G, "rate_copy : Hz")
        S.connect(j="i")
        S.run_regularly("rate_copy = rates_pre")
        run(defaultclock.dt)
    assert_allclose(S.rate_copy[:], np.arange(len(G)) * 42 * Hz)


@pytest.mark.cpp_standalone
@pytest.mark.standalone_only
def test_negative_duration_in_standalone_device():
    set_device("cpp_standalone")
    G = NeuronGroup(1, "v:1")
    with pytest.raises(ValueError):
        run(-1 * second)


if __name__ == "__main__":
    for t in [
        test_cpp_standalone,
        test_multiple_connects,
        test_storing_loading,
        test_openmp_consistency,
        test_duplicate_names_across_nets,
        test_openmp_scalar_writes,
        test_time_after_run,
        test_array_cache,
        test_run_with_debug,
        test_changing_profile_arg,
        test_profile_via_set_device_arg,
        test_delete_code_data,
        test_delete_directory,
        test_multiple_standalone_runs,
        test_change_parameter_without_recompile,
        test_change_parameter_without_recompile_errors,
        test_change_parameter_without_recompile_dict_syntax,
        test_change_parameter_without_recompile_dependencies,
        test_change_synapse_parameter_without_recompile_dict_syntax,
        test_change_parameters_multiprocessing,
        test_header_file_inclusion,
    ]:
        t()
        reinit_and_delete()



{# USES_VARIABLES { N, t, rate, _clock_t, _clock_dt, _spikespace,
                    _num_source_neurons, _source_start, _source_stop } #}
{% extends 'common_group.pyx' %}

{% block maincode %}

    cdef size_t _num_spikes = {{_spikespace}}[_num{{_spikespace}}-1]
    
    # For subgroups, we do not want to record all spikes
    # We assume that spikes are ordered
    cdef int _start_idx = -1
    cdef int _end_idx = -1
    cdef size_t _j
    for _j in range(_num_spikes):
        _idx = {{_spikespace}}[_j]
        if _idx >= _source_start:
            _start_idx = _j
            break
    if _start_idx == -1:
        _start_idx = _num_spikes
    for _j in range(_start_idx, _num_spikes):
        _idx = {{_spikespace}}[_j]
        if _idx >= _source_stop:
            _end_idx = _j
            break
    if _end_idx == -1:
        _end_idx =_num_spikes
    _num_spikes = _end_idx - _start_idx
    
    # Calculate the new length for the arrays
    cdef size_t _new_len = {{_dynamic_t}}.shape[0] + 1

    # Resize the arrays
    _owner.resize(_new_len)
    {{N}} = _new_len

    # Set the new values
    {{_dynamic_t}}.data[_new_len-1] = {{_clock_t}}
    {{_dynamic_rate}}.data[_new_len-1] = _num_spikes/{{_clock_dt}}/_num_source_neurons

{% endblock %}



"""
An example of turning an expensive `Synapses.connect` operation into
three cheap ones using a mathematical trick.

Consider the connection probability between neurons i and j given by
the Gaussian function :math:`p=e^{-\alpha(i-j)^2}` (for some constant
:math:`\alpha`). If we want to connect neurons with this probability,
we can very simply do::

    S.connect(p='exp(-alpha*(i-j)**2)')

However, this has a problem. Although we know that this will create
:math:`O(N)` synapses if N is the number of neurons, because we
have specified ``p`` as a function of i and j, we have to evaluate
``p(i, j)`` for every pair ``(i, j)``, and therefore it takes
:math:`O(N^2)` operations.

Our first option is to take a cutoff, and say that if :math:`p<q` for some
small :math:`q`, then we assume that :math:`p\approx 0`. We can work out
which j values are compatible with a given value of i by solving
:math:`e^{-\alpha(i-j)^2}<q` which gives
:math:`|i-j|<\sqrt{-\log(q)/\alpha)}=w`. Now we implement the rule
using the generator syntax to only search for values between ``i-w``
and ``i+w``, except that some of these values will be outside the
valid range of values for j so we set ``skip_if_invalid=True``.
The connection code is then::

    S.connect(j='k for k in range(i-w, i+w) if rand()<exp(-alpha*(i-j)**2)',
              skip_if_invalid=True)

This is a lot faster (see graph labelled "Limited" for this algorithm).

However, it may be a problem that we have to specify a cutoff and so
we will lose some synapses doing this: it won't be mathematically exact.
This isn't a problem for the Gaussian because w grows very slowly with
the cutoff probability q, but for other probability distributions with
more weight in the tails, it could be an issue.

If we want to be exact, we can still do a big improvement. For the
case :math:`i-w\leq j\leq i+w` we use the same connection code, but
we also handle the case :math:`|i-j|>w`. This time, we note that we
want to create a synapse with probability :math:`p(i-j)` and we can
rewrite this as :math:`p(i-j)/p(w)\cdot p(w)`. If :math:`|i-j|>w`
then this is a product of two probabilities :math:`p(i-j)/p(w)`
and :math:`p(w)`. So in the region :math:`|i-j|>w` a synapse
will be created if two random events both occur, with these
two probabilities. This might seem a little strange until you
notice that one of the two probabilities :math:`p(w)` doesn't
depend on i or j. This lets us use the much more efficient
``sample`` algorithm to generate a set of candidate ``j`` values,
and then add the additional test ``rand()<p(i-j)/p(w)``. Here's the
code for that::

    w = int(ceil(sqrt(log(q)/-0.1)))
    S.connect(j='k for k in range(i-w, i+w) if rand()<exp(-alpha*(i-j)**2)',
              skip_if_invalid=True)
    pmax = exp(-0.1*w**2)
    S.connect(j='k for k in sample(0, i-w, p=pmax) if rand()<exp(-alpha*(i-j)**2)/pmax',
              skip_if_invalid=True)
    S.connect(j='k for k in sample(i+w, N_post, p=pmax) if rand()<exp(-alpha*(i-j)**2)/pmax',
              skip_if_invalid=True)

This "Divided" method is also much faster than the naive method,
and is mathematically correct. Note though that this method is still
:math:`O(N^2)` but the constants are much, much smaller and this
will usually be sufficient. It is possible to take the ideas
developed here even further and get even better scaling, but in
most cases it's unlikely to be worth the effort.

The code below shows these examples written out, along with
some timing code and plots for different values of N.
"""
from brian2 import *
import time

def naive(N):
    G = NeuronGroup(N, 'v:1', threshold='v>1', name='G')
    S = Synapses(G, G, on_pre='v += 1', name='S')
    S.connect(p='exp(-0.1*(i-j)**2)')

def limited(N, q=0.001):
    G = NeuronGroup(N, 'v:1', threshold='v>1', name='G')
    S = Synapses(G, G, on_pre='v += 1', name='S')
    w = int(ceil(sqrt(log(q)/-0.1)))
    S.connect(j='k for k in range(i-w, i+w) if rand()<exp(-0.1*(i-j)**2)', skip_if_invalid=True)

def divided(N, q=0.001):
    G = NeuronGroup(N, 'v:1', threshold='v>1', name='G')
    S = Synapses(G, G, on_pre='v += 1', name='S')
    w = int(ceil(sqrt(log(q)/-0.1)))
    S.connect(j='k for k in range(i-w, i+w) if rand()<exp(-0.1*(i-j)**2)', skip_if_invalid=True)
    pmax = exp(-0.1*w**2)
    S.connect(j='k for k in sample(0, i-w, p=pmax) if rand()<exp(-0.1*(i-j)**2)/pmax', skip_if_invalid=True)
    S.connect(j='k for k in sample(i+w, N_post, p=pmax) if rand()<exp(-0.1*(i-j)**2)/pmax', skip_if_invalid=True)

def repeated_run(f, N, repeats):
    start_time = time.time()
    for _ in range(repeats):
        f(N)
    end_time = time.time()
    return (end_time-start_time)/repeats

N = array([100, 500, 1000, 5000, 10000, 20000])
repeats = array([100, 10, 10, 1, 1, 1])*3
naive(10)
limited(10)
divided(10)
print('Starting naive')
loglog(N, [repeated_run(naive, n, r) for n, r in zip(N, repeats)],
       label='Naive', lw=2)
print('Starting limit')
loglog(N, [repeated_run(limited, n, r) for n, r in zip(N, repeats)],
       label='Limited', lw=2)
print('Starting divided')
loglog(N, [repeated_run(divided, n, r) for n, r in zip(N, repeats)],
       label='Divided', lw=2)
xlabel('N')
ylabel('Time (s)')
legend(loc='best', frameon=False)
show()



"""
This example shows how to run several, independent simulations in standalone mode using multiple processes to run the
simulations in parallel.
Given that this example only involves a single neuron, an alternative – and arguably more elegant – solution
would be to run the simulations in a single `NeuronGroup`, where each neuron receives input with a different rate.

The example is a standalone equivalent of the one presented in :doc:`/tutorials/3-intro-to-brian-simulations`.

Note that Python's `multiprocessing` module cannot deal with user-defined functions (including `TimedArray`) and other
complex code structures. If you run into `PicklingError` or `AttributeError` exceptions, you might
have to use the `pathos` (https://pypi.org/project/pathos) package instead, which can handle more complex
code structures.
"""
import numpy as np
import matplotlib.pyplot as plt
import brian2 as b2
from time import time

b2.set_device('cpp_standalone', build_on_run=False)

class SimWrapper:
    def __init__(self):
        self.net = b2.Network()
        P = b2.PoissonGroup(num_inputs, rates=input_rate)
        eqs = """
            dv/dt = -v/tau : 1
            tau : second (constant)
            """
        G = b2.NeuronGroup(1, eqs, threshold='v>1', reset='v=0', method='euler', name='neuron')
        S = b2.Synapses(P, G, on_pre='v += weight')
        S.connect()
        M = b2.SpikeMonitor(G, name='spike_monitor')
        self.net.add([P, G, S, M])

        self.net.run(1000 * b2.ms)

        self.device = b2.get_device()
        self.device.build(run=False, directory=None)  # compile the code, but don't run it yet

    def do_run(self, tau_i):
        # Workaround to set the device globally in this context
        from brian2.devices import device_module
        device_module.active_device = self.device

        result_dir = f'result_{tau_i}'
        self.device.run(run_args={self.net['neuron'].tau: tau_i},
                        results_directory=result_dir)
        return self.net["spike_monitor"].num_spikes/ b2.second


if __name__ == "__main__":
    start_time = time()
    num_inputs = 100
    input_rate = 10 * b2.Hz
    weight = 0.1

    npoints = 15
    tau_range = np.linspace(1, 15, npoints) * b2.ms    

    sim = SimWrapper()

    from multiprocessing import Pool
    with Pool(npoints) as pool:
        output_rates = pool.map(sim.do_run, tau_range)

    print(f"Done in {time() - start_time}")

    plt.plot(tau_range/b2.ms, output_rates)
    plt.xlabel(r"$\tau$ (ms)")
    plt.ylabel("Firing rate (sp/s)")
    plt.show()



import inspect
import pydoc
import re
import textwrap

from sphinx.pycode import ModuleAnalyzer

from .docscrape import ClassDoc, FunctionDoc, NumpyDocString


class SphinxDocString(NumpyDocString):
    def __init__(self, docstring, config=None):
        if config is None:
            config = {}
        NumpyDocString.__init__(self, docstring, config=config)

    # string conversion routines
    @staticmethod
    def _str_header(name, symbol="`"):
        return [f".. rubric:: {name}", ""]

    @staticmethod
    def _str_field_list(name):
        return [f":{name}:"]

    @staticmethod
    def _str_indent(doc, indent=4):
        out = []
        for line in doc:
            out += [" " * indent + line]
        return out

    def _str_summary(self):
        return self["Summary"] + [""]

    def _str_extended_summary(self):
        return self["Extended Summary"] + [""]

    def _str_param_list(self, name):
        out = []
        if self[name]:
            out += self._str_field_list(name)
            out += [""]
            for param, param_type, desc in self[name]:
                out += self._str_indent([f"**{param.strip()}** : {param_type}"])
                out += [""]
                out += self._str_indent(desc, 8)
                out += [""]

        return out

    @property
    def _obj(self):
        if hasattr(self, "_cls"):
            return self._cls
        elif hasattr(self, "_f"):
            return self._f
        return None

    def _str_member_list(self):
        """
        Generate a member listing, autosummary:: table .

        """
        out = []

        for name in ["Attributes", "Methods"]:
            if not self[name]:
                continue

            out += [f".. rubric:: {name}", ""]
            prefix = getattr(self, "_name", "")

            if prefix:
                prefix = f"{prefix}."

            autosum = []
            for param, _, desc in self[name]:
                param = param.strip()
                if self._obj:
                    # Fake the attribute as a class property, but do not touch
                    # methods
                    if hasattr(self._obj, "__module__") and not (
                        hasattr(self._obj, param)
                        and callable(getattr(self._obj, param))
                    ):
                        # Do not override directly provided docstrings
                        if not len("".join(desc).strip()):
                            analyzer = ModuleAnalyzer.for_module(self._obj.__module__)
                            desc = analyzer.find_attr_docs().get(
                                (self._obj.__name__, param), ""
                            )

                        # Only fake a property if we got a docstring
                        if len("".join(desc).strip()):
                            setattr(
                                self._obj,
                                param,
                                property(lambda self: None, doc="\n".join(desc)),
                            )

                if len(prefix):
                    autosum += [f"   ~{prefix}{param}"]
                else:
                    autosum += [f"   {param}"]

            if autosum:
                out += [".. autosummary::", ""]
                out += autosum

            out += [""]
        return out

    def _str_member_docs(self, name):
        """
        Generate the full member autodocs

        """
        out = []

        if self[name]:
            prefix = getattr(self, "_name", "")

            if prefix:
                prefix += "."

            for param, _, _ in self[name]:
                if name == "Methods":
                    out += [f".. automethod:: {prefix}{param}"]
                elif name == "Attributes":
                    out += [f".. autoattribute:: {prefix}{param}"]

            out += [""]
        return out

    def _str_section(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            out += [""]
            content = textwrap.dedent("\n".join(self[name])).split("\n")
            out += content
            out += [""]
        return out

    def _str_see_also(self, func_role):
        out = []
        if self["See Also"]:
            see_also = super()._str_see_also(func_role)
            out = [".. seealso::", ""]
            out += self._str_indent(see_also[2:])
        return out

    def _str_raises(self, name, func_role):
        if not self[name]:
            return []
        out = []
        out += self._str_header(name)
        for func, _, desc in self[name]:
            out += [f":exc:`{func}`"]
            if desc:
                out += self._str_indent([" ".join(desc)])
        out += [""]
        return out

    def _str_warnings(self):
        out = []
        if self["Warnings"]:
            out = [".. warning::", ""]
            out += self._str_indent(self["Warnings"])
        return out

    def _str_index(self):
        idx = self["index"]
        out = []
        if len(idx) == 0:
            return out

        out += [f".. index:: {idx.get('default', '')}"]
        for section, references in idx.items():
            if section == "default":
                continue
            elif section == "refguide":
                out += [f"   single: {', '.join(references)}"]
            else:
                out += [f"   {section}: {','.join(references)}"]
        return out

    def _str_references(self):
        out = []
        if self["References"]:
            out += self._str_header("References")
            if isinstance(self["References"], str):
                self["References"] = [self["References"]]
            out.extend(self["References"])
            out += [""]
            # Latex collects all references to a separate bibliography,
            # so we need to insert links to it
            out += [".. only:: latex", ""]
            items = []
            for line in self["References"]:
                m = re.match(r".. \[([a-z0-9._-]+)\]", line, re.I)
                if m:
                    items.append(m.group(1))
            out += [f"   {', '.join([f'[{item}]_' for item in items])}", ""]
        return out

    def _str_examples(self):
        return self._str_section("Examples")

    def __str__(self, indent=0, func_role="brianobj"):
        out = []
        out += self._str_index() + [""]
        out += self._str_summary()
        out += self._str_extended_summary()
        for param_list in ("Parameters", "Returns", "Other Parameters"):
            out += self._str_param_list(param_list)
        for param_list in ("Raises", "Warns"):
            out += self._str_raises(param_list, func_role)
        out += self._str_warnings()
        out += self._str_see_also(func_role)
        out += self._str_section("Notes")
        out += self._str_references()
        out += self._str_examples()
        out += self._str_member_list()
        if self["Attributes"] + self["Methods"]:
            out += [".. rubric:: Details", ""]
            for param_list in ("Attributes", "Methods"):
                out += self._str_member_docs(param_list)
        out = self._str_indent(out, indent)
        return "\n".join(out)


class SphinxFunctionDoc(SphinxDocString, FunctionDoc):
    def __init__(self, obj, doc=None, config=None):
        if config is None:
            config = {}
        FunctionDoc.__init__(self, obj, doc=doc, config=config)


class SphinxClassDoc(SphinxDocString, ClassDoc):
    def __init__(self, obj, doc=None, func_doc=None, name=None, config=None):
        if config is None:
            config = {}
        self.name = name
        ClassDoc.__init__(self, obj, doc=doc, func_doc=None, config=config)


class SphinxObjDoc(SphinxDocString):
    def __init__(self, obj, doc=None, config=None):
        if config is None:
            config = {}
        self._f = obj
        SphinxDocString.__init__(self, doc, config=config)


def get_doc_object(obj, what=None, doc=None, name=None, config=None):
    if config is None:
        config = {}
    if what is None:
        if inspect.isclass(obj):
            what = "class"
        elif inspect.ismodule(obj):
            what = "module"
        elif callable(obj):
            what = "function"
        else:
            what = "object"
    if what == "class":
        return SphinxClassDoc(
            obj, func_doc=SphinxFunctionDoc, doc=doc, name=name, config=config
        )
    elif what in ("function", "method"):
        return SphinxFunctionDoc(obj, doc=doc, config=config)
    else:
        if doc is None:
            doc = pydoc.getdoc(obj)
        return SphinxObjDoc(obj, doc, config=config)



import os
import sys
import warnings
import runpy
import pytest

from brian2 import device, set_device

import numpy as np
warnings.simplefilter('ignore')


class ExampleRun(pytest.Item):
    '''
    A test case that simply executes a python script
    '''
    @classmethod
    def from_parent(cls, filename, codegen_target, dtype, parent):
        super_class = super(ExampleRun, cls)
        if hasattr(super_class, 'from_parent'):
            new_node = super_class.from_parent(parent=parent,
                                               name=ExampleRun.id(filename))
        else:
            # For pytest < 6
            new_node = cls(parent=parent, name=ExampleRun.id(filename))
        new_node.filename = filename
        new_node.codegen_target = codegen_target
        new_node.dtype = dtype
        return new_node

    @staticmethod
    def id(filename):
        # Remove the .py and pretend the dirname is a package and the filename
        # is a class.
        name = os.path.splitext(os.path.split(filename)[1])[0]
        pkgname = os.path.split(os.path.split(filename)[0])[1]
        return pkgname + '.' + name.replace('.', '_')

    def shortDescription(self):
        return str(self)

    def runtest(self):
        import matplotlib as _mpl
        _mpl.use('Agg')
        import numpy as np
        from brian2 import prefs
        from brian2.utils.filetools import ensure_directory_of_file
        prefs.codegen.target = self.codegen_target
        prefs.core.default_float_dtype = self.dtype
        # Move to the file's directory for the run, so that it can do relative
        # imports and load files (e.g. figure style files)
        curdir = os.getcwd()
        os.chdir(os.path.dirname(self.filename))
        sys.path.append(os.path.dirname(self.filename))
        try:
            runpy.run_path(self.filename, run_name='__main__')
            if self.codegen_target == 'cython' and self.dtype == np.float64:
                for fignum in _mpl.pyplot.get_fignums():
                    fname = os.path.relpath(self.filename, self.example_dir)
                    fname = fname.replace('/', '.').replace('\\\\', '.')
                    fname = fname.replace('.py', '.%d.png' % fignum)
                    fname = os.path.abspath(os.path.join(self.example_dir,
                                                         '../docs_sphinx/resources/examples_images/',
                                                         fname))
                    ensure_directory_of_file(fname)
                    _mpl.pyplot.figure(fignum).savefig(fname)
        finally:
            _mpl.pyplot.close('all')
            os.chdir(curdir)
            sys.path.remove(os.path.dirname(self.filename))
            device.reinit()
            set_device('runtime')

    def __str__(self):
        return 'Example: %s (%s, %s)' % (self.filename, self.codegen_target,
                                         self.dtype.__name__)


class ExampleCollector(pytest.Collector):
    @classmethod
    def from_parent(cls, example_dir, parent):
        collector = super(ExampleCollector, cls)
        if hasattr(collector, 'from_parent'):
            new_collector = collector.from_parent(parent,
                                                  name='example_collector')
        else:
            # For pytest < 6
            new_collector = cls(parent=parent, name='example_collector')
        new_collector.example_dir = example_dir
        return new_collector

    def collect(self):
        for dirname, dirs, files in os.walk(self.example_dir):
            for filename in files:
                if filename.endswith('.py'):
                    run = ExampleRun.from_parent(os.path.join(dirname, filename),
                                                 'cython',
                                                 np.float64,
                                                 parent=self)
                    run.example_dir = self.example_dir
                    yield run


class Plugin:
    def __init__(self, example_dir):
        self.example_dir = example_dir

    def pytest_collect_file(self, path, parent):
        return ExampleCollector.from_parent(self.example_dir, parent=parent)


if __name__ == '__main__':
    example_dir = os.path.abspath(os.path.join(__file__, '..', '..', '..', 'examples'))
    sys.exit(pytest.main([__file__, '--verbose'],
                         plugins=[Plugin(example_dir)]))



# cython: language_level = 3
# distutils: language = c++
# distutils: sources = brian2/synapses/cspikequeue.cpp

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.string cimport string

import cython
from cython.operator import dereference
from cython.operator cimport dereference

cimport numpy as np
import numpy as np

np.import_array()

cdef extern from "stdint_compat.h":
    # Longness only used for type promotion
    # Actual compile time size used for conversion
    ctypedef signed int int32_t
    ctypedef signed long int64_t

ctypedef pair[int, vector[vector[int32_t]]] state_pair

ctypedef fused float_array:
    np.ndarray[double, ndim=1, mode='c']
    np.ndarray[float, ndim=1, mode='c']

cdef extern from "cspikequeue.cpp":
    cdef cppclass CSpikeQueue:
        CSpikeQueue(int, int) except +
        void prepare[scalar](scalar*, int, int32_t*, int, double)
        void push(int32_t *, int)
        state_pair _full_state()
        void _restore_from_full_state(state_pair)
        vector[int32_t]* peek()
        void advance()

cdef class SpikeQueue:
    # TODO: Currently, the data type for dt and delays is fixed
    cdef CSpikeQueue *thisptr
    cdef readonly tuple _state_tuple

    def __cinit__(self, int source_start, int source_end):
        self.thisptr = new CSpikeQueue(source_start, source_end)

    def __init__(self, source_start, source_end):
        self._state_tuple = (source_start, source_end, np.int32)

    def __dealloc__(self):
        del self.thisptr

    def _full_state(self):
        return self.thisptr._full_state()

    cdef object __weakref__  # Allows weak references to the SpikeQueue

    def _restore_from_full_state(self, state):
        cdef vector[vector[int32_t]] empty_queue
        cdef state_pair empty_state
        if state is not None:
            self.thisptr._restore_from_full_state(state)
        else:
            empty_queue = vector[vector[int32_t]]()
            empty_state = (0, empty_queue)
            self.thisptr._restore_from_full_state(empty_state)

    def prepare(self, float_array real_delays,
                double dt,
                np.ndarray[int32_t, ndim=1, mode='c'] sources):
        if real_delays.dtype == np.float32:
            self.thisptr.prepare(<float*>real_delays.data,
                                 real_delays.shape[0],
                                 <int32_t*>sources.data,
                                 sources.shape[0], dt)
        else:
            self.thisptr.prepare(<double*>real_delays.data,
                                 real_delays.shape[0],
                                 <int32_t*>sources.data,
                                 sources.shape[0], dt)

    def push(self, np.ndarray[int32_t, ndim=1, mode='c'] spikes):
        self.thisptr.push(<int32_t*>spikes.data, spikes.shape[0])

    def peek(self):
        # will only be used if the queue has size > 0
        cdef int32_t* spikes_data
        cdef np.npy_intp shape[1]
        # This should create a numpy array from a std::vector<int> without
        # copying -- &spikes[0] is guaranteed to point to a contiguous array
        # according to the C++ standard.
        cdef vector[int32_t]* spikes = self.thisptr.peek()
        cdef size_t spikes_size = dereference(spikes).size()
        if spikes_size == 0:
            return np.empty(0, dtype=np.int32)
        else:
            spikes_data = <int32_t*>(&(dereference(spikes)[0]))
            shape[0] = spikes_size
            return np.PyArray_SimpleNewFromData(1, shape, np.NPY_INT32, spikes_data)

    def advance(self):
        self.thisptr.advance()



import fnmatch
import glob
import os
import shutil
from collections import defaultdict


class GlobDirectoryWalker:
    # a forward iterator that traverses a directory tree

    def __init__(self, directory, pattern="*"):
        self.stack = [directory]
        self.pattern = pattern
        self.files = []
        self.index = 0

    def __getitem__(self, index):
        while True:
            try:
                file = self.files[self.index]
                self.index = self.index + 1
            except IndexError:
                # pop next directory from stack
                self.directory = self.stack.pop()
                if os.path.isdir(self.directory):
                    self.files = os.listdir(self.directory)
                else:
                    self.files = []
                self.index = 0
            else:
                # got a filename
                fullname = os.path.join(self.directory, file)
                if os.path.isdir(fullname) and not os.path.islink(fullname):
                    self.stack.append(fullname)
                if fnmatch.fnmatch(file, self.pattern):
                    return fullname


def main(rootpath, destdir):
    if not os.path.exists(destdir):
        shutil.os.makedirs(destdir)

    examplesfnames = [fname for fname in GlobDirectoryWalker(rootpath, "*.py")]
    allowed_additional_file_extensions = [".rst", ".md", ".txt", ".mplstyle"]
    additional_files = [
        fname
        for fname in GlobDirectoryWalker(rootpath, "*.*")
        if os.path.splitext(fname)[1] in allowed_additional_file_extensions
    ]

    data_file_extensions = [".npy"]
    data_files = [
        fname
        for fname in GlobDirectoryWalker(rootpath, "*.*")
        if os.path.splitext(fname)[1] in data_file_extensions
    ]

    print(f"Documenting {len(examplesfnames)} examples")
    print(f"Documenting {len(additional_files)} additional files")
    print(f"Documenting {len(data_files)} data files")

    examplespaths = []
    examplesbasenames = []
    relativepaths = []
    outnames = []
    for f in examplesfnames:
        path, file = os.path.split(f)
        relpath = os.path.relpath(path, rootpath)
        if relpath == ".":
            relpath = ""
        path = os.path.normpath(path)
        filebase, ext = os.path.splitext(file)
        exname = filebase
        if relpath:
            exname = relpath.replace("/", ".").replace("\\", ".") + "." + exname
        examplespaths.append(path)
        examplesbasenames.append(filebase)
        relativepaths.append(relpath)
        outnames.append(exname)
    # We assume all files are encoded as UTF-8
    examplescode = []
    for fname in examplesfnames:
        with open(fname, encoding="utf-8") as f:
            examplescode.append(f.read())
    examplesdocs = []
    examplesafterdoccode = []
    for code in examplescode:
        codesplit = code.split("\n")
        comment_lines = 0
        for line in codesplit:
            if line.startswith("#") or len(line) == 0:
                comment_lines += 1
            else:
                break
        codesplit = codesplit[comment_lines:]
        readingdoc = False
        doc = []
        afterdoccode = ""
        for i in range(len(codesplit)):
            stripped = codesplit[i].strip()
            if stripped[:3] == '"""' or stripped[:3] == "'''":
                if not readingdoc:
                    readingdoc = True
                else:
                    afterdoccode = "\n".join(codesplit[i + 1 :])
                    break
            elif readingdoc:
                doc.append(codesplit[i])
            else:  # No doc
                afterdoccode = "\n".join(codesplit[i:])
                break
        examplesdocs.append("\n".join(doc))
        examplesafterdoccode.append(afterdoccode)

    categories = defaultdict(list)
    examples = zip(
        examplesfnames,
        examplespaths,
        examplesbasenames,
        examplescode,
        examplesdocs,
        examplesafterdoccode,
        relativepaths,
        outnames,
    )
    # Get the path relative to the examples director (not relative to the
    # directory where this file is installed
    if "BRIAN2_DOCS_EXAMPLE_DIR" in os.environ:
        rootdir = os.environ["BRIAN2_DOCS_EXAMPLE_DIR"]
    else:
        rootdir, _ = os.path.split(__file__)
        rootdir = os.path.normpath(os.path.join(rootdir, "../../examples"))
    eximgpath = os.path.abspath(
        os.path.join(rootdir, "../docs_sphinx/resources/examples_images")
    )
    print("Searching for example images in directory", eximgpath)
    for _fname, _path, basename, _code, docs, afterdoccode, relpath, exname in examples:
        categories[relpath].append((exname, basename))
        title = "Example: " + basename
        output = ".. currentmodule:: brian2\n\n"
        output += ".. " + basename + ":\n\n"
        output += title + "\n" + "=" * len(title) + "\n\n"
        note = f"""
        .. only:: html

            .. |launchbinder| image:: http://mybinder.org/badge.svg
            .. _launchbinder: https://mybinder.org/v2/gh/brian-team/brian2-binder/master?filepath=examples/{exname.replace('.', '/')}.ipynb

            .. note::
               You can launch an interactive, editable version of this
               example without installing any local files
               using the Binder service (although note that at some times this
               may be slow or fail to open): |launchbinder|_

        """
        output += note + "\n\n"
        output += docs + "\n\n::\n\n"
        output += "\n".join(["    " + line for line in afterdoccode.split("\n")])
        output += "\n\n"

        eximgpattern = os.path.join(eximgpath, f"{exname}.*")
        images = glob.glob(eximgpattern + ".png") + glob.glob(eximgpattern + ".gif")
        for image in sorted(images):
            _, image = os.path.split(image)
            print("Found example image file", image)
            output += f".. image:: ../resources/examples_images/{image}\n\n"

        with open(os.path.join(destdir, exname + ".rst"), "w", encoding="utf-8") as f:
            f.write(output)

    category_additional_files = defaultdict(list)
    for fname in additional_files:
        path, file = os.path.split(fname)
        relpath = os.path.relpath(path, rootpath)
        if relpath == ".":
            relpath = ""
        full_name = relpath.replace("/", ".").replace("\\", ".") + "." + file + ".rst"
        category_additional_files[relpath].append((file, full_name))
        with open(fname, encoding="utf-8") as f:
            content = f.read()
        output = file + "\n" + "=" * len(file) + "\n\n"
        output += ".. code-block:: none\n\n"
        content_lines = ["\t" + line for line in content.split("\n")]
        output += "\n".join(content_lines)
        output += "\n\n"
        with open(os.path.join(destdir, full_name), "w", encoding="utf-8") as f:
            f.write(output)

    category_data_files = defaultdict(list)
    for fname in data_files:
        path, file = os.path.split(fname)
        relpath = os.path.relpath(path, rootpath)
        if relpath == ".":
            relpath = ""
        full_name = relpath.replace("/", ".").replace("\\", ".") + "." + file + ".rst"
        category_data_files[relpath].append((file, full_name))
        output = file + "\n" + "=" * len(file) + "\n\n"
        output += (
            "Download:"
            f" :download:`{file} <{os.path.join('/../examples/', relpath, file)}>`\n\n"
        )
        with open(os.path.join(destdir, full_name), "w", encoding="utf-8") as f:
            f.write(output)

    mainpage_text = "Examples\n"
    mainpage_text += "========\n\n"

    def insert_category(category, mainpage_text):
        if category:
            label = category.lower().replace(" ", "-").replace("/", ".")
            mainpage_text += f"\n.. _{label}:\n\n"
            mainpage_text += "\n" + category + "\n" + "-" * len(category) + "\n\n"
        mainpage_text += ".. toctree::\n"
        mainpage_text += "   :maxdepth: 1\n\n"
        for exname, basename in sorted(categories[category]):
            mainpage_text += f"   {basename} <{exname}>\n"
        for fname, full_name in sorted(category_additional_files[category]):
            mainpage_text += f"   {fname} <{full_name}>\n"
        for fname, full_name in sorted(category_data_files[category]):
            mainpage_text += f"   {fname} <{full_name}>\n"
        return mainpage_text

    mainpage_text = insert_category("", mainpage_text)
    for category in sorted(categories.keys()):
        if category:
            mainpage_text = insert_category(category, mainpage_text)

    with open(os.path.join(destdir, "index.rst"), "w") as f:
        f.write(mainpage_text)


if __name__ == "__main__":
    main("../../examples", "../../docs_sphinx/examples")



"""
Demonstrate the correctness of the "derivative-free Milstein method" for
multiplicative noise.
"""
from brian2 import *
# We only get exactly the same random numbers for the exact solution and the
# simulation if we use the numpy code generation target
prefs.codegen.target = 'numpy'

# setting a random seed makes all variants use exactly the same Wiener process
seed = 12347  

X0 = 1
mu = 0.5/second # drift
sigma = 0.1/second #diffusion 

runtime = 1*second


def simulate(method, dt):
    """
    simulate geometrical Brownian with the given method
    """
    np.random.seed(seed)
    G = NeuronGroup(1, 'dX/dt = (mu - 0.5*second*sigma**2)*X + X*sigma*xi*second**.5: 1',
                    dt=dt, method=method)
    G.X = X0
    mon = StateMonitor(G, 'X', record=True)
    net = Network(G, mon)
    net.run(runtime)
    return mon.t_[:], mon.X.flatten()


def exact_solution(t, dt):
    """
    Return the exact solution for geometrical Brownian motion at the given
    time points
    """
    # Remove units for simplicity
    my_mu = float(mu)
    my_sigma = float(sigma)
    dt = float(dt)
    t = asarray(t)
    
    np.random.seed(seed)
    # We are calculating the values at the *start* of a time step, as when using
    # a StateMonitor. Therefore the Brownian motion starts with zero
    brownian = np.hstack([0, cumsum(sqrt(dt) * np.random.randn(len(t)-1))])
    
    return (X0 * exp((my_mu - 0.5*my_sigma**2)*(t+dt) + my_sigma*brownian))

figure(1, figsize=(16, 7))
figure(2, figsize=(16, 7))

methods = ['milstein', 'heun']
dts = [1*ms, 0.5*ms, 0.2*ms, 0.1*ms, 0.05*ms, 0.025*ms, 0.01*ms, 0.005*ms]

rows = int(sqrt(len(dts)))
cols = int(ceil(1.0 * len(dts) / rows))
errors = dict([(method, zeros(len(dts))) for method in methods])
for dt_idx, dt in enumerate(dts):
    print('dt: %s' % dt)
    trajectories = {}
    # Test the numerical methods
    for method in methods:
        t, trajectories[method] = simulate(method, dt)
    # Calculate the exact solution
    exact = exact_solution(t, dt)    
    
    for method in methods:
        # plot the trajectories
        figure(1)
        subplot(rows, cols, dt_idx+1)
        plot(t, trajectories[method], label=method, alpha=0.75)
        
        # determine the mean absolute error
        errors[method][dt_idx] = mean(abs(trajectories[method] - exact))
        # plot the difference to the real trajectory
        figure(2)
        subplot(rows, cols, dt_idx+1)
        plot(t, trajectories[method] - exact, label=method, alpha=0.75)
        
    figure(1)
    plot(t, exact, color='gray', lw=2, label='exact', alpha=0.75)
    title('dt = %s' % str(dt))
    xticks([])

figure(1)
legend(frameon=False, loc='best')
tight_layout()

figure(2)
legend(frameon=False, loc='best')
tight_layout()

figure(3)
for method in methods:
    plot(array(dts) / ms, errors[method], 'o', label=method)
legend(frameon=False, loc='best')
xscale('log')
yscale('log')
xlabel('dt (ms)')
ylabel('Mean absolute error')
tight_layout()

show()



# coding=utf-8
"""
Modeling neuron-glia interactions with the Brian 2 simulator
Marcel Stimberg, Dan F. M. Goodman, Romain Brette, Maurizio De Pittà
bioRxiv 198366; doi: https://doi.org/10.1101/198366

Figure 4C: Closed-loop gliotransmission.

I/O curves in terms average per-spike release vs. rate of stimulation for three
synapses: one without gliotransmission, and the other two with open- and close-loop
gliotransmssion.
"""
from brian2 import *

import plot_utils as pu

set_device('cpp_standalone', directory=None)  # Use fast "C++ standalone mode"
seed(1929)  # to get identical figures for repeated runs

################################################################################
# Model parameters
################################################################################
### General parameters
N_synapses = 100
N_astro = 2
transient = 15*second
duration = transient + 180*second  # Total simulation time
sim_dt = 1*ms                      # Integrator/sampling step

### Neuron parameters

# ### Synapse parameters
### Synapse parameters
rho_c = 0.005               # Synaptic vesicle-to-extracellular space volume ratio
Y_T = 500*mmolar            # Total vesicular neurotransmitter concentration
Omega_c = 40/second         # Neurotransmitter clearance rate
U_0__star = 0.6             # Resting synaptic release probability
Omega_f = 3.33/second       # Synaptic facilitation rate
Omega_d = 2.0/second        # Synaptic depression rate
# --- Presynaptic receptors
O_G = 1.5/umolar/second     # Agonist binding (activating) rate
Omega_G = 0.5/(60*second)   # Agonist release (deactivating) rate

### Astrocyte parameters
# ---  Calcium fluxes
O_P = 0.9*umolar/second     # Maximal Ca^2+ uptake rate by SERCAs
K_P = 0.05 * umolar         # Ca2+ affinity of SERCAs
C_T = 2*umolar              # Total cell free Ca^2+ content
rho_A = 0.18                # ER-to-cytoplasm volume ratio
Omega_C = 6/second          # Maximal rate of Ca^2+ release by IP_3Rs
Omega_L = 0.1/second        # Maximal rate of Ca^2+ leak from the ER
# --- IP_3R kinectics
d_1 = 0.13*umolar           # IP_3 binding affinity
d_2 = 1.05*umolar           # Ca^2+ inactivation dissociation constant
O_2 = 0.2/umolar/second     # IP_3R binding rate for Ca^2+ inhibition
d_3 = 0.9434*umolar         # IP_3 dissociation constant
d_5 = 0.08*umolar           # Ca^2+ activation dissociation constant
# --- IP_3 production
# --- Agonist-dependent IP_3 production
O_beta = 3.2*umolar/second  # Maximal rate of IP_3 production by PLCbeta
O_N = 0.3/umolar/second     # Agonist binding rate
Omega_N = 0.5/second        # Maximal inactivation rate
K_KC = 0.5*umolar           # Ca^2+ affinity of PKC
zeta = 10                   # Maximal reduction of receptor affinity by PKC
# --- Endogenous IP3 production
O_delta = 0.6*umolar/second # Maximal rate of IP_3 production by PLCdelta
kappa_delta = 1.5* umolar   # Inhibition constant of PLC_delta by IP_3
K_delta = 0.1*umolar        # Ca^2+ affinity of PLCdelta
# --- IP_3 degradation
Omega_5P = 0.05/second      # Maximal rate of IP_3 degradation by IP-5P
K_D = 0.7*umolar            # Ca^2+ affinity of IP3-3K
K_3K = 1.0*umolar           # IP_3 affinity of IP_3-3K
O_3K = 4.5*umolar/second    # Maximal rate of IP_3 degradation by IP_3-3K
# --- IP_3 diffusion
F_ex = 2.0*umolar/second    # Maximal exogenous IP3 flow
I_Theta = 0.3*umolar        # Threshold gradient for IP_3 diffusion
omega_I = 0.05*umolar       # Scaling factor of diffusion
# --- Gliotransmitter release and time course
C_Theta = 0.5*umolar        # Ca^2+ threshold for exocytosis
Omega_A = 0.6/second        # Gliotransmitter recycling rate
U_A = 0.6                   # Gliotransmitter release probability
G_T = 200*mmolar            # Total vesicular gliotransmitter concentration
rho_e = 6.5e-4              # Astrocytic vesicle-to-extracellular volume ratio
Omega_e = 60/second         # Gliotransmitter clearance rate
alpha = 0.0                 # Gliotransmission nature

################################################################################
# Model definition
################################################################################
defaultclock.dt = sim_dt  # Set the integration time

f_vals = np.logspace(-1, 2, N_synapses)*Hz
source_neurons = PoissonGroup(N_synapses, rates=f_vals)
target_neurons = NeuronGroup(N_synapses, '')

### Synapses
# Note that the synapse does not actually have any effect on the post-synaptic
# target
# Also note that for easier plotting we do not use the "event-driven" flag here,
# even though the value of u_S and x_S only needs to be updated on the arrival
# of a spike
synapses_eqs = '''
# Neurotransmitter
dY_S/dt = -Omega_c * Y_S : mmolar (clock-driven)
# Fraction of activated presynaptic receptors
dGamma_S/dt = O_G * G_A * (1 - Gamma_S) - Omega_G * Gamma_S : 1 (clock-driven)
# Usage of releasable neurotransmitter per single action potential:
du_S/dt = -Omega_f * u_S : 1 (event-driven)
# Fraction of synaptic neurotransmitter resources available for release:
dx_S/dt = Omega_d *(1 - x_S) : 1 (event-driven)
r_S : 1  # released synaptic neurotransmitter resources
G_A : mmolar  # gliotransmitter concentration in the extracellular space
'''
synapses_action = '''
U_0 = (1 - Gamma_S) * U_0__star + alpha * Gamma_S
u_S += U_0 * (1 - u_S)
r_S = u_S * x_S
x_S -= r_S
Y_S += rho_c * Y_T * r_S
'''
synapses = Synapses(source_neurons, target_neurons,
                    model=synapses_eqs, on_pre=synapses_action,
                    method='exact')
# We create three synapses per connection: only the first two are modulated by
# the astrocyte however. Note that we could also create three synapses per
# connection with a single connect call by using connect(j='i', n=3), but this
# would create synapses arranged differently (synapses connection pairs
# (0, 0), (0, 0), (0, 0), (1, 1), (1, 1), (1, 1), ..., instead of
# connections (0, 0), (1, 1), ..., (0, 0), (1, 1), ..., (0, 0), (1, 1), ...)
# making the later connection descriptions more complicated.
synapses.connect(j='i')  # closed-loop modulation
synapses.connect(j='i')  # open modulation
synapses.connect(j='i')  # no modulation
synapses.x_S = 1.0

### Astrocytes
# The astrocyte emits gliotransmitter when its Ca^2+ concentration crosses
# a threshold
astro_eqs = '''
# Fraction of activated astrocyte receptors:
dGamma_A/dt = O_N * Y_S * (1 - Gamma_A) -
              Omega_N*(1 + zeta * C/(C + K_KC)) * Gamma_A : 1

# IP_3 dynamics:
dI/dt = J_beta + J_delta - J_3K - J_5P + J_ex             : mmolar
J_beta = O_beta * Gamma_A                                 : mmolar/second
J_delta = O_delta/(1 + I/kappa_delta) *
                         C**2/(C**2 + K_delta**2)         : mmolar/second
J_3K = O_3K * C**4/(C**4 + K_D**4) * I/(I + K_3K)         : mmolar/second
J_5P = Omega_5P*I                                         : mmolar/second
delta_I_bias = I - I_bias : mmolar
J_ex = -F_ex/2*(1 + tanh((abs(delta_I_bias) - I_Theta)/omega_I)) *
                sign(delta_I_bias)                        : mmolar/second
I_bias                                                    : mmolar (constant)

# Ca^2+-induced Ca^2+ release:
dC/dt = (Omega_C * m_inf**3 * h**3 + Omega_L) * (C_T - (1 + rho_A)*C) -
        O_P * C**2/(C**2 + K_P**2) : mmolar
dh/dt = (h_inf - h)/tau_h          : 1  # IP3R de-inactivation probability
m_inf = I/(I + d_1) * C/(C + d_5)  : 1
h_inf = Q_2/(Q_2 + C)              : 1
tau_h = 1/(O_2 * (Q_2 + C))        : second
Q_2 = d_2 * (I + d_1)/(I + d_3)    : mmolar

# Fraction of gliotransmitter resources available for release
dx_A/dt = Omega_A * (1 - x_A) : 1
# gliotransmitter concentration in the extracellular space
dG_A/dt = -Omega_e*G_A        : mmolar
# Neurotransmitter concentration in the extracellular space
Y_S                           : mmolar
'''
glio_release = '''
G_A += rho_e * G_T * U_A * x_A
x_A -= U_A *  x_A
'''
astrocyte = NeuronGroup(N_astro*N_synapses, astro_eqs,
                        # The following formulation makes sure that a "spike" is
                        # only triggered at the first threshold crossing
                        threshold='C>C_Theta',
                        refractory='C>C_Theta',
                        # The gliotransmitter release happens when the threshold
                        # is crossed, in Brian terms it can therefore be
                        # considered a "reset"
                        reset=glio_release,
                        method='rk4')
astrocyte.h = 0.9
astrocyte.x_A = 1.0
# Only the second group of N_synapses astrocytes are activated by external stimulation
astrocyte.I_bias = (np.r_[np.zeros(N_synapses), np.ones(N_synapses)])*1.0*umolar

## Connections
ecs_syn_to_astro = Synapses(synapses, astrocyte,
                            'Y_S_post = Y_S_pre : mmolar (summed)')
# Connect the first N_synapses synapses--astrocyte pairs
ecs_syn_to_astro.connect(j='i if i < N_synapses')

ecs_astro_to_syn = Synapses(astrocyte, synapses,
                            'G_A_post = G_A_pre : mmolar (summed)')
# Connect the first N_synapses astrocytes--pairs
# (closed-loop configuration)
ecs_astro_to_syn.connect(j='i if i < N_synapses')
# Connect the second N_synapses astrocyte--synapses pairs
# (open-loop configuration)
ecs_astro_to_syn.connect(j='i if i >= N_synapses and i < 2*N_synapses')

################################################################################
# Monitors
################################################################################
syn_mon = StateMonitor(synapses, 'r_S',
                       record=np.arange(N_synapses*(N_astro+1)))

################################################################################
# Simulation run
################################################################################
run(duration, report='text')

################################################################################
# Analysis and plotting
################################################################################
plt.style.use('figures.mplstyle')

fig, ax = plt.subplots(nrows=4, ncols=1, figsize=(3.07, 3.07*1.33), sharex=False,
                       gridspec_kw={'height_ratios': [1, 3, 3, 3],
                                    'top': 0.98, 'bottom': 0.12,
                                    'left': 0.22, 'right': 0.93})

## Turn off one axis to display accordingly to the other figure in example_4_synrel.py
ax[0].axis('off')

ax[1].errorbar(f_vals/Hz, np.mean(syn_mon.r_S[2*N_synapses:], axis=1),
               np.std(syn_mon.r_S[2*N_synapses:], axis=1),
               fmt='o', color='black', lw=0.5)
ax[1].set(xlim=(0.08, 100), xscale='log',
          ylim=(0., 0.7),
          ylabel=r'$\langle r_S \rangle$')
pu.adjust_spines(ax[1], ['left'])

ax[2].errorbar(f_vals/Hz, np.mean(syn_mon.r_S[N_synapses:2*N_synapses], axis=1),
               np.std(syn_mon.r_S[N_synapses:2*N_synapses], axis=1),
               fmt='o', color='C2', lw=0.5)
ax[2].set(xlim=(0.08, 100), xscale='log',
          ylim=(0., 0.2), ylabel=r'$\langle r_S \rangle$')
pu.adjust_spines(ax[2], ['left'])

ax[3].errorbar(f_vals/Hz, np.mean(syn_mon.r_S[:N_synapses], axis=1),
               np.std(syn_mon.r_S[:N_synapses], axis=1),
               fmt='o', color='C3', lw=0.5)
ax[3].set(xlim=(0.08, 100), xticks=np.logspace(-1, 2, 4), xscale='log',
          ylim=(0., 0.7), xlabel='input frequency (Hz)',
          ylabel=r'$\langle r_S \rangle$')
ax[3].xaxis.set_major_formatter(ScalarFormatter())
pu.adjust_spines(ax[3], ['left', 'bottom'])

pu.adjust_ylabels(ax, x_offset=-0.2)

plt.show()